from .model import model
from .libinterferometry import Visibilities
import numpy

def center(data, params):
    
    if type(params) == list:
        params = numpy.array(params)
    elif type(params) == numpy.ndarray:
        pass
    
    centering_params = [params[0], params[1], 1.]
    
    data_complex = data.real+1j*data.imag

    model_complex = numpy.empty(data.real.shape, dtype=complex)
    for i in range(len(data.freq)):
        model_complex[:,i] = model(data.u*data.freq[i]/data.freq.mean(), \
                data.v*data.freq[i]/data.freq.mean(), centering_params, \
                funct="point", return_type="complex")[:,0]
    
    centered_data = data_complex * model_complex.conj()
    
    return Visibilities(data.u.copy(), data.v.copy(), data.freq.copy(), \
            centered_data.real, centered_data.imag, data.weights.copy())



def control(incl_dust=None, incl_lines=None, incl_freefree=None, \
        nphot_therm=None, nphot_scat=None, nphot_spec=None, iseed=None, \
        ifast=None, enthres=None, itempdecoup=None, istar_sphere=None, \
        ntemp=None, temp0=None, temp1=None, scattering_mode_max=None, \
        rto_style=None, camera_tracemode=None, camera_nrrefine=None, \
        camera_refine_criterion=None, camera_incl_stars=None, \
        camera_starsphere_nrpix=None, camera_spher_cavity_relres=None, \
        camera_localobs_projection=None, camera_min_dangle=None, \
        camera_max_dangle=None, camera_min_dr=None, \
        camera_diagnostics_subpix=None, camera_secondorder=None, \
        camera_interpol_jnu=None, mc_weighted_photons=None, \
        optimized_motion=None, lines_mode=None, lines_maxdoppler=None, \
        lines_partition_ntempint=None, lines_partition_temp0=None, \
        lines_partition_temp1=None, lines_show_pictograms=None, \
        tgas_eq_tdust=None, subbox_nx=None, subbox_ny=None, subbox_nz=None, \
        subbox_x0=None, subbox_x1=None, subbox_y0=None, subbox_y1=None, \
        subbox_z0=None, subbox_z1=None, modified_random_walk=None, \
        mrw_gamma=None, mrw_tauthres=None, mrw_count_trigger=None, \
        setthreads=1, camera_scatsrc_allfreq=None, mc_scat_maxtauabs=None, \
        writeimage_unformatted=None):

    f = open("radmc3d.inp","w")

    if (incl_dust != None):
        f.write("incl_dust = {0:d}\n".format(incl_dust))
    if (incl_lines != None):
        f.write("incl_lines = {0:d}\n".format(incl_lines))
    if (incl_freefree != None):
        f.write("incl_freefree = {0:d}\n".format(incl_freefree))
    if (nphot_therm != None):
        f.write("nphot_therm = {0:d}\n".format(int(nphot_therm)))
    if (nphot_scat != None):
        f.write("nphot_scat = {0:d}\n".format(int(nphot_scat)))
    if (nphot_spec != None):
        f.write("nphot_spec = {0:d}\n".format(int(nphot_spec)))
    if (iseed != None):
        f.write("iseed = {0:d}\n".format(iseed))
    if (ifast != None):
        f.write("ifast = {0:d}\n".format(ifast))
    if (enthres != None):
        f.write("enthres = {0:f}\n".format(enthres))
    if (itempdecoup != None):
        f.write("itempdecoup = {0:d}\n".format(itempdecoup))
    if (istar_sphere != None):
        f.write("istar_sphere = {0:d}\n".format(istar_sphere))
    if (ntemp != None):
        f.write("ntemp = {0:d}\n".format(ntemp))
    if (temp0 != None):
        f.write("temp0 = {0:f}\n".format(temp0))
    if (temp1 != None):
        f.write("temp1 = {0:f}\n".format(temp1))
    if (scattering_mode_max != None):
        f.write("scattering_mode_max = {0:d}\n".format(scattering_mode_max))
    if (rto_style != None):
        f.write("rto_style = {0:d}\n".format(rto_style))
    if (camera_tracemode != None):
        f.write("camera_tracemode = {0:d}\n".format(camera_tracemode))
    if (camera_nrrefine != None):
        f.write("camera_nrrefine = {0:d}\n".format(camera_nrrefine))
    if (camera_refine_criterion != None):
        f.write("camera_refine_criterion = {0:f}\n".format( \
                camera_refine_criterion))
    if (camera_incl_stars != None):
        f.write("camera_incl_stars = {0:d}\n".format(camera_incl_stars))
    if (camera_starsphere_nrpix != None):
        f.write("camera_starsphere_nrpix = {0:d}\n".format( \
                camera_starsphere_nrpix))
    if (camera_spher_cavity_relres != None):
        f.write("camera_spher_cavity_relres = {0:f}\n".format( \
                camera_spher_cavity_relres))
    if (camera_localobs_projection != None):
        f.write("camera_localobs_projection = {0:d}\n".format( \
                camera_localobs_projection))
    if (camera_min_dangle != None):
        f.write("camera_min_dangle = {0:f}\n".format(camera_min_dangle))
    if (camera_max_dangle != None):
        f.write("camera_max_dangle = {0:f}\n".format(camera_max_dangle))
    if (camera_min_dr != None):
        f.write("camera_min_dr = {0:f}\n".format(camera_min_dr))
    if (camera_diagnostics_subpix != None):
        f.write("camera_diagnostics_subpix = {0:d}\n".format( \
                camera_diagnostics_subpix))
    if (camera_secondorder != None):
        f.write("camera_secondorder = {0:d}\n".format(camera_secondorder))
    if (camera_interpol_jnu != None):
        f.write("camera_interpol_jnu = {0:d}\n".format(camera_interpol_jnu))
    if (camera_scatsrc_allfreq != None):
        f.write("camera_scatsrc_allfreq = {0:d}\n".format(camera_scatsrc_allfreq))
    if (mc_weighted_photons != None):
        f.write("mc_weighted_photons = {0:d}\n".format(mc_weighted_photons))
    if (optimized_motion != None):
        f.write("optimized_motion = {0:d}\n".format(optimized_motion))
    if (lines_mode != None):
        f.write("lines_mode = {0:d}\n".format(lines_mode))
    if (lines_maxdoppler != None):
        f.write("lines_maxdoppler = {0:f}\n".format(lines_maxdoppler))
    if (lines_partition_ntempint != None):
        f.write("lines_partition_ntempint = {0:d}\n".format( \
                lines_partition_ntempint))
    if (lines_partition_temp0 != None):
        f.write("lines_partition_temp0 = {0:f}\n".format(lines_partition_temp0))
    if (lines_partition_temp1 != None):
        f.write("lines_partition_temp1 = {0:f}\n".format(lines_partition_temp1))
    if (lines_show_pictograms != None):
        f.write("lines_show_pictograms = {0:d}\n".format(lines_show_pictograms))
    if (tgas_eq_tdust != None):
        f.write("tgas_eq_tdust = {0:d}\n".format(tgas_eq_tdust))
    if (subbox_nx != None):
        f.write("subbox_nx = {0:d}\n".format(subbox_nx))
    if (subbox_ny != None):
        f.write("subbox_ny = {0:d}\n".format(subbox_ny))
    if (subbox_nz != None):
        f.write("subbox_nz = {0:d}\n".format(subbox_nz))
    if (subbox_x0 != None):
        f.write("subbox_x0 = {0:d}\n".format(subbox_x0))
    if (subbox_x1 != None):
        f.write("subbox_x1 = {0:d}\n".format(subbox_x1))
    if (subbox_y0 != None):
        f.write("subbox_y0 = {0:d}\n".format(subbox_y0))
    if (subbox_y1 != None):
        f.write("subbox_y1 = {0:d}\n".format(subbox_y1))
    if (subbox_z0 != None):
        f.write("subbox_z0 = {0:d}\n".format(subbox_z0))
    if (subbox_z1 != None):
        f.write("subbox_z1 = {0:d}\n".format(subbox_z1))
    if (modified_random_walk != None):
        f.write("modified_random_walk = {0:d}\n".format(modified_random_walk))
    if (mrw_gamma != None):
        f.write("mrw_gamma = {0:d}\n".format(mrw_gamma))
    if (mrw_count_trigger != None):
        f.write("mrw_count_trigger = {0:d}\n".format(mrw_count_trigger))
    if (mrw_tauthres != None):
        f.write("mrw_tauthres = {0:d}\n".format(mrw_tauthres))
    if (setthreads != 1):
        f.write("setthreads = {0:d}\n".format(setthreads))
    if (mc_scat_maxtauabs != None):
        f.write("mc_scat_maxtauabs = {0:d}\n".format(mc_scat_maxtauabs))
    if (writeimage_unformatted != None):
        f.write("writeimage_unformatted = {0:d}\n".format(writeimage_unformatted))

    f.close()

def stars(rstar, mstar, lam, xstar, ystar, zstar, tstar=None, fstar=None):

    nstars = len(rstar)
    nlam = len(lam)

    f = open("stars.inp","w")

    f.write(str(2)+"\n")
    f.write("{0:d}  {1:d}\n".format(nstars, nlam))

    for istar in range (nstars):
        f.write("{0:e}   {1:e}   {2:e}   {3:e}   {4:e}\n".format(rstar[istar], \
                mstar[istar], xstar[istar], ystar[istar], zstar[istar]))

    for ilam in range(nlam):
        f.write("{0:e}\n".format(lam[ilam]))

    for istar in range(nstars):
        if (tstar[istar] != 0):
            f.write("{0:f}\n".format(-tstar[istar]))
        else:
            for i in range(nlam):
                f.write("{0:e}\n".format(fstar[ilam]))

    f.close()

def wavelength_micron(lam):

    nlam = len(lam)

    f = open("wavelength_micron.inp","w")

    f.write("{0:d}\n".format(nlam))
    for ilam in range(nlam):
        f.write("{0:f}\n".format(lam[ilam]))

    f.close()

def camera_wavelength_micron(lam):

    nlam = len(lam)

    f = open("camera_wavelength_micron.inp","w")

    f.write("{0:d}\n".format(nlam))
    for ilam in range(nlam):
        f.write("{0:f}\n".format(lam[ilam]))

    f.close()

def amr_grid(x, y, z, gridstyle="regular", coordsystem="cartesian"):

    nx = x.size-1
    ny = y.size-1
    nz = z.size-1

    incl_x = int(nx > 1)
    incl_y = int(ny > 1)
    incl_z = int(nz > 1)

    f = open("amr_grid.inp","w")

    f.write(str(1)+"\n")

    if (gridstyle == "regular"):
        f.write("0\n")
    elif (gridstyle == "octtree"):
        f.write("1\n")
    elif (gridstyle == "amr"):
        f.write("10\n")

    if (coordsystem == "cartesian"):
        f.write("0\n")
    elif (coordsystem == "spherical"):
        f.write("100\n")
    elif (coordsystem == "cylindrical"):
        f.write("200\n")

    f.write("0\n")
    f.write("{0:d}  {1:d}  {2:d}\n".format(incl_x, incl_y, incl_z))
    f.write("{0:d}  {1:d}  {2:d}\n".format(nx, ny, nz))

    if (gridstyle == "octtree"):
        print("OctTree grids not yet implemented.")
    elif (gridstyle == "amr"):
        print("Layer-style AMR grids not yet implemented.")

    for i in range(nx+1):
        f.write("{0:12.9e}\n".format(x[i]))
    for i in range(ny+1):
        f.write("{0:12.9e}\n".format(y[i]))
    for i in range(nz+1):
        f.write("{0:12.9e}\n".format(z[i]))

    # Insert extra info for octtree and amr grids here...

    f.close()

def dust_density(density, gridstyle="normal"):

    nspecies = len(density)

    if (gridstyle == "normal"):
        nx, ny, nz = density[0].shape
        ncells = nx*ny*nz

    f = open("dust_density.inp","w")
    f.write("1\n")
    f.write("{0:d}\n".format(ncells))
    f.write("{0:d}\n".format(nspecies))

    for ispec in range(nspecies):
        if (gridstyle == "normal"):
            for iz in range(nz):
                for iy in range(ny):
                    for ix in range(nx):
                        f.write("{0:e}\n".format(density[ispec][ix,iy,iz]))

    f.close()

def dust_temperature(temperature, gridstyle="normal"):

    nspecies = len(temperature)

    if (gridstyle == "normal"):
        nx, ny, nz = temperature[0].shape
        ncells = nx*ny*nz

    f = open("dust_temperature.dat","w")
    f.write("1\n")
    f.write("{0:d}\n".format(ncells))
    f.write("{0:d}\n".format(nspecies))

    for ispec in range(nspecies):
        if (gridstyle == "normal"):
            for iz in range(nz):
                for iy in range(ny):
                    for ix in range(nx):
                        f.write("{0:f}\n".format(temperature[ispec][ix,iy,iz]))

    f.close()

def dustopac(opacity):

    nspecies = len(opacity)

    f = open("dustopac.inp","w")

    f.write("2\n")
    f.write("{0:d}\n".format(nspecies))
    f.write("==============================================================\n")
    for i in range(nspecies):
        filetype = opacity[i].split("_")[0]
        species = opacity[i].split("_")[1].split(".")[0]

        if (filetype == "dustkappa"):
            f.write("1\n")
        elif (filetype == "dustkapscatmat"):
            f.write("10\n")
        elif (filetype == "dustopac"):
            f.write("-1\n")

        f.write("0\n")
        f.write("{0:s}\n".format(species))
        f.write("----------------------------------------------------------\n")

    f.close()

def dustkappa(species, lam, kabs, ksca=None, g=None):

    if (type(ksca) != type(None)) and (type(g) != type(None)):
        iformat = 3
    elif (type(ksca) != type(None)) and (type(g) == type(None)):
        iformat = 2
    else:
        iformat = 1

    nlam = len(lam)

    f = open("dustkappa_{0:s}.inp".format(species),"w")

    f.write("{0:d}\n".format(iformat))
    f.write("{0:d}\n".format(nlam))

    for ilam in range(nlam):
        if (iformat == 1):
            f.write("{0:e}   {1:e}\n".format(lam[ilam],kabs[ilam]))
        elif (iformat == 2):
            f.write("{0:e}   {1:e}   {2:e}\n".format(lam[ilam], kabs[ilam], \
                    ksca[ilam]))
        elif (iformat == 3):
            f.write("{0:e}   {1:e}   {2:e}   {3:e}\n".format(lam[ilam], \
                    kabs[ilam], ksca[ilam], g[ilam]))

    f.close()

def line(species, inpstyle, colpartners):

    nspecies = len(species)

    f = open("lines.inp", "w")

    f.write("2\n")
    f.write("{0:d}\n".format(nspecies))

    for i in range(nspecies):
        f.write("{0:s} {1:s} 0 0 {2:d}\n".format(species[i], inpstyle[i], \
                len(colpartners[i])))
        if (len(colpartners[i]) > 0):
            for j in range(len(colpartners[i])):
                f.write("{0:a}\n".format(colpartners[i][j]))

    f.close()

def molecule(species, name):

    f = open("molecule_{0:s}.inp".format(name), "w")

    f.write("!MOLECULE\n")
    f.write("{0:s}\n".format(name))
    f.write("!MOLECULAR WEIGHT\n")
    f.write("{0:.1f}\n".format(species.mass))
    f.write("!NUMBER OF ENERGY LEVELS\n")
    f.write("{0:d}\n".format(species.J.size))
    f.write("!LEVEL + ENERGIES(cm^-1) + WEIGHT + J\n")

    for i in range(species.J.size):
        f.write("{0:>5d}{1:>16.9f}{2:>10.1f}{3:>10s}\n".format(i+1, \
                species.E[i], species.g[i], species.J[i]))

    f.write("!NUMBER OF RADIATIVE TRANSITIONS\n")
    f.write("{0:d}\n".format(species.J_u.size))
    f.write("!TRANS + UP + LOW + EINSTEINA(s^-1) + FREQ(GHz) + E_u(K)\n")

    for i in range(species.J_u.size):
        f.write("{0:>5d}{1:>6d}{2:>6d}{3:>12.3e}{4:>16.7f}{5:>10.2f}\n".format(\
                i+1, int(species.J_u[i]), int(species.J_l[i]), \
                species.A_ul[i], species.nu[i]/1.0e9, species.E_u[i]))

    f.write("!NUMBER OF COLL PARTNERS\n")
    f.write("{0:d}\n".format(len(species.partners)))
    
    for i in range(len(species.partners)):
        f.write("!COLLISIONS BETWEEN\n")
        f.write("{0:s}".format(species.partners[i]))
        f.write("!NUMBER OF COLL TRANS\n")
        f.write("{0:d}\n".format(species.J_u_coll[i].size))
        f.write("!NUMBER OF COLL TEMPS\n")
        f.write("{0:d}\n".format(species.temp[i].size))
        f.write("!COLL TEMPS\n")
        temp_string = ""
        for j in range(species.temp[i].size):
            temp_string += "{0:>7.1f}".format(species.temp[i][j])
        f.write(temp_string+"\n")
        f.write("!TRANS + UP + LOW + COLLRATES(cm^3 s^-1)\n")
        for j in range(species.J_u_coll[i].size):
            string = "{0:>5d}{1:>6d}{2:>6d} ".format(j+1, \
                    int(species.J_u_coll[i][j]), int(species.J_l_coll[i][j]))
            for k in range(species.temp[i].size):
                string += "{0:10.3e}".format(species.gamma[i][j,k])
            f.write(string+"\n")
        f.write("\n")

    f.close()

def numberdens(n, species, gridstyle="normal"):

    if (gridstyle == "normal"):
        nx, ny, nz = n.shape
        ncells = nx*ny*nz

    f = open("numberdens_{0:s}.inp".format(species),"w")
    f.write("1\n")
    f.write("{0:d}\n".format(ncells))

    if (gridstyle == "normal"):
        for iz in range(nz):
            for iy in range(ny):
                for ix in range(nx):
                    f.write("{0:e}\n".format(n[ix,iy,iz]))

    f.close()

def gas_temperature(T, gridstyle="normal"):

    if (gridstyle == "normal"):
        nx, ny, nz = T.shape
        ncells = nx*ny*nz

    f = open("gas_temperature.inp","w")
    f.write("1\n")
    f.write("{0:d}\n".format(ncells))

    if (gridstyle == "normal"):
        for iz in range(nz):
            for iy in range(ny):
                for ix in range(nx):
                    f.write("{0:e}\n".format(T[ix,iy,iz]))

    f.close()

def microturbulence(a_turb, gridstyle="normal"):

    if (gridstyle == "normal"):
        nx, ny, nz = a_turb.shape
        ncells = nx*ny*nz

    f = open("microturbulence.inp","w")
    f.write("1\n")
    f.write("{0:d}\n".format(ncells))

    if (gridstyle == "normal"):
        for iz in range(nz):
            for iy in range(ny):
                for ix in range(nx):
                    f.write("{0:e}\n".format(a_turb[ix,iy,iz]))

    f.close()

def gas_velocity(v, gridstyle="normal"):

    if (gridstyle == "normal"):
        nx, ny, nz = v[0].shape
        ncells = nx*ny*nz

    f = open("gas_velocity.inp", "w")
    f.write("1\n")
    f.write("{0:d}\n".format(ncells))

    if (gridstyle == "normal"):
        for iz in range(nz):
            for iy in range(ny):
                for ix in range(nx):
                    f.write("{0:e} {1:e} {2:e}\n".format(v[0][ix,iy,iz], \
                            v[1][ix,iy,iz], v[2][ix,iy,iz]))

    f.close()

def scattering_phase(scattering_phase, freq, gridstyle="normal"):

    nfreq = freq.size

    if (gridstyle == "normal"):
        nx, ny, nz = scattering_phase[0].shape
        ncells = nx*ny*nz

    f = open("scattering_phase.out","w")
    f.write("2\n")
    f.write("{0:d}\n".format(ncells))
    f.write("{0:d}\n".format(nspecies))

    f.write("       ".join(freq)+"\n")

    for ifreq in range(nfreq):
        if (gridstyle == "normal"):
            for iz in range(nz):
                for iy in range(ny):
                    for ix in range(nx):
                        f.write("{0:e}\n".format(density[ifreq,ix,iy,iz]))

    f.close()



#!/usr/bin/env python

from scipy.integrate import trapz
from pdspy.constants.astronomy import AU, M_sun
from pdspy.constants.physics import c
import pdspy.interferometry as uv
import pdspy.modeling as modeling
import pdspy.utils as utils
import scipy.integrate
import emcee
import numpy
import glob
import sys
import os

# First check what version of the code was run.

if len(glob.glob("chain.npy")) > 0:
    version = "emcee"
elif len(glob.glob("results.hdf5")) > 0:
    version = "emcee3"
elif len(glob.glob("samples.npy")) > 0 or len(glob.glob("sampler.p")) > 0:
    version = "dynesty"
else:
    version = "none"
    print("Did not detect any pdspy results... just upgrading config.py and data.")

if version != "none":
    print("Detected results from the {0:s} version of pdspy.".format(version))

if version == "dynesty":
    print("** Please note that because of the way that dynesty is run, it is impossible to update sampler.p. This means you will not be able to resume an in-progress fit, nor will you be able to re-run disk_model_nested.py, flared_model_nested, etc. within this model directory and get correct results. The resulting samples.npy will be updated so that you can still work with the results, if they exist. Otherwise, you will need to re-run the fit. **")

# Now load in the configuration file.

config = utils.load_config()

# Determine what files are going to be changed.

print("Checking which files need to be updated...")

changed_files = ["config.py"]

for f in config.visibilities["file"]:
    if os.path.exists("/".join(f.split("/")[0:-1]+[".upgraded_to_pdspy2"])):
        print(f+" has already been upgraded... skipping.")
    else:
        changed_files.append(f)

if version == "emcee":
    if os.path.exists("chain.npy"):
        changed_files += ["chain.npy","pos.npy","prob.npy"]
elif version == "emcee3":
    if os.path.exists("results.hdf5"):
        changed_files += ["results.hdf5"]
elif version == "dynesty":
    if os.path.exists("samples.npy"):
        changed_files += ["samples.npy"]
    if os.path.exists("sampler.p"):
        changed_files += ["sampler.p"]

print("Permanent changes will be made to the following files:")
for f in changed_files:
    print("    {0:s}".format(f))
result = input("Proceed? (y/n)")

if result == "n":
    sys.exit(0)

# Check whether this directory was already converted.

if os.path.exists(".upgraded_to_pdspy2"):
    print("Based on the presence of a .upgraded_to_pdspy2 file in this directory, it seems that this directory has already been converted to v2. Re-running the code is dangerous and would likely lead to incorrect results. If you do really want to re-run, please restore the files in this directory to their original state and then delete the .upgraded_to_pdspy file.")
    sys.exit(0)
else:
    f = open(".upgraded_to_pdspy2","w")
    f.write("upgraded_to_pdspy2\n")
    f.close()

# Update the visibility data to take the complex conjugate.

for j, f in enumerate(config.visibilities["file"]):
    if not os.path.exists("/".join(f.split("/")[0:-1]+[".upgraded_to_pdspy2"])):
        print("Updating "+f)

        data = uv.Visibilities()
        data.read(f)

        data.imag *= -1.

        # Check whether there's a difference in config.visibilities["lam"][j]
        # and the frequency of the data.

        if data.freq.size == 1:
            data.freq[0] = c / (float(config.visibilities["lam"][j])*1.0e-4)
            
            model = "disk"
        else:
            model = "flared"

        # Write out the updated data file.

        os.system("mv "+f+" "+f+".backup")

        data.write(f)

        u = open("/".join(f.split("/")[0:-1]+[".upgraded_to_pdspy2"]),"w")
        u.write("upgraded_to_pdspy2\n")
        u.close()

# Finally, update config.py to have the appropriate numbers.

print("Updating config.py")

os.system("cp config.py config.py.backup")

f = open("config.py","r")
lines = f.readlines()
f.close()

new_lines = []
for i, line in enumerate(lines):
    if "x0" in line or "y0" in line:
        val = line[line.find("[")+1 : line.find("]")]
        vals = val.split(",")
        new_vals = [str(float(v)*-1) for v in vals]
        if "fixed" in line and len(vals) == 2:
            new_vals = new_vals[::-1]
        line = line.replace(val, ",".join(new_vals))
    elif '"pa"' in line:
        val = line[line.find("[")+1 : line.find("]")]
        vals = val.split(",")
        new_vals = [str(-float(v)) for v in vals][::-1]
        line = line.replace(val, ",".join(new_vals))
    new_lines.append(line)

f = open("config.py","w")
for line in new_lines:
    f.write(line)
f.close()

# First, if the results are emcee < 3, convert to emcee3.

if version == "emcee" and os.path.exists("chain.npy"):
    print("Converting to emcee v3")

    # Load in the current information.

    chain = numpy.load("chain.npy")
    prob = numpy.load("prob.npy")

    nwalkers = chain.shape[0]
    nsteps = chain.shape[1]
    ndim = chain.shape[2]

    # Create an artificial random state.

    state = numpy.random.mtrand.RandomState().get_state()

    # Adjust prob to be the same shape as the chain.

    if prob.shape[1] < chain.shape[1]:
        new_prob = numpy.zeros((chain.shape[0],chain.shape[1]))
        new_prob[:,-prob.shape[1]:] = prob.copy()
        prob = new_prob.copy()

    # Create an emcee backend.

    backend = emcee.backends.HDFBackend("results.hdf5")
    backend.reset(nwalkers, ndim)

    backend.grow(nsteps, None)

    # Now loop through the steps and save to the backend.

    for i in range(nsteps):
        sample = emcee.State(chain[:,i,:], log_prob=prob[:,i], \
                random_state=state)
        accepted = numpy.ones((nwalkers,))

        backend.save_step(sample, accepted)

    # Finally, move the old results to .backup files.

    os.system("mv chain.npy chain.npy.backup")
    os.system("mv prob.npy prob.npy.backup")
    os.system("mv pos.npy pos.npy.backup")

    version = "emcee3"

# Now, change the values of tne appropriate parameters in the results.

if os.path.exists("samples.npy") or os.path.exists("results.hdf5"):
    if version == "dynesty":
        samples = numpy.load("samples.npy")
    elif version == "emcee3":
        results = emcee.backends.HDFBackend("results.hdf5")
        chain = results.get_chain()
        prob = results.get_log_prob()

    # Get the keys of parameters and the index of particular ones that need to 
    # be updated.

    keys = []
    for key in sorted(config.parameters.keys()):
        if not config.parameters[key]["fixed"]:
            keys.append(key)
    ndim = len(keys)

    for ikey in range(ndim):
        if keys[ikey] == "x0":
            ix0 = ikey
        elif keys[ikey] == "y0":
            iy0 = ikey
        elif keys[ikey] == "pa":
            ipa = ikey
        elif keys[ikey] == "logM_disk":
            iMdisk = ikey
        elif keys[ikey] == "logM_env":
            iMenv = ikey

    # Define a function that calculates the correct masses.

    def calculate_correct_masses(params, disk=True, envelope=True):
        # Generate the model.

        m = modeling.run_disk_model(config.visibilities, config.images, \
                config.spectra, params, config.parameters, plot=False, \
                ncpus=4, ncpus_highmass=4, source="V883Ori", \
                no_radiative_transfer=True)

        rho, theta, phi = numpy.meshgrid(m.grid.r*AU, m.grid.theta, \
                m.grid.phi, indexing='ij')

        # Update the disk density by calculating the unnormalized density, i.e. 
        # how we used to do it, and integrating to get the normalization.

        if disk:
            r_high = numpy.logspace(numpy.log10(m.disk.rmin), \
                    numpy.log10(10*m.disk.rmax), 1000)
            Sigma_high = m.disk.surface_density(r_high, normalize=False)

            disk_mass = 2*numpy.pi*trapz(r_high*AU*Sigma_high, r_high*AU)
        else:
            disk_mass = None

        # Update the envelope mass as well.

        if envelope:
            env_density_new = m.envelope.density(m.grid.r, m.grid.theta, \
                    m.grid.phi)

            m.envelope.cavrfact = 1.0
            env_density_old = m.envelope.density(m.grid.r, m.grid.theta, \
                    m.grid.phi)

            scale = (env_density_old / env_density_new).min()

            envelope_density = scale * env_density_new

            env_mass = (4*numpy.pi*scipy.integrate.trapz(scipy.integrate.trapz(\
                    envelope_density*rho**2*numpy.sin(theta), theta, axis=1), \
                    rho[:,0,:],axis=0))[0]
        else:
            env_mass = None

        return disk_mass, env_mass

    # Now do the conversions.

    if version == "dynesty":
        print("Updating samples.npy")

        for i in range(samples.shape[0]):
            if "x0" in keys:
                samples[i,ix0] *= -1
            if "y0" in keys:
                samples[i,iy0] *= -1
            if "pa" in keys:
                samples[i,ipa] *= -1
                if model == "disk":
                    samples[i,ipa] += 180

            params = dict(zip(keys,samples[i,:]))
            disk_mass, env_mass = calculate_correct_masses(params, \
                    disk="logM_disk" in keys, envelope="logM_env" in keys)

            if "logM_disk" in keys:
                samples[i,iMdisk] = numpy.log10(disk_mass/M_sun)
            if "logM_env" in keys:
                samples[i,iMenv] = numpy.log10(env_mass/M_sun)

        os.system("cp samples.npy samples.npy.backup")
        numpy.save("samples.npy", samples)

    elif version == "emcee3":
        print("Updating results.hdf5")

        for i in range(chain.shape[0]):
            for j in range(chain.shape[1]):
                # Fix x0, y0, and pa.

                if "x0" in keys:
                    chain[i,j,ix0] *= -1
                if "y0" in keys:
                    chain[i,j,iy0] *= -1
                if "pa" in keys:
                    chain[i,j,ipa] *= -1
                    if model == "disk":
                        chain[i,j,ipa] += 180

                params = dict(zip(keys,chain[i,j,:]))
                disk_mass, env_mass = calculate_correct_masses(params, \
                        disk="logM_disk" in keys, envelope="logM_env" in keys)

                if "logM_disk" in keys:
                    chain[i,j,iMdisk] = numpy.log10(disk_mass/M_sun)
                if "logM_env" in keys:
                    chain[i,j,iMenv] = numpy.log10(env_mass/M_sun)

        # Save the updated chain.

        os.system("cp results.hdf5 results.hdf5.backup")

        nsteps = chain.shape[0]
        nwalkers = chain.shape[1]
        ndim = chain.shape[2]

        state = numpy.random.mtrand.RandomState().get_state()

        backend.reset(nwalkers, ndim)
        backend.grow(nsteps, None)

        for i in range(nsteps):
            sample = emcee.State(chain[i], log_prob=prob[i], random_state=state)
            accepted = numpy.ones((nwalkers,))

            backend.save_step(sample, accepted)

# If sampler.p exists, move it to backup so you can't resume a fit.

if version == "dynesty" and os.path.exists("sampler.p"):
    print("Moving sampler.p to sampler.p.backup so the dynesty fit cannot be resumed.")
    os.system("mv sampler.p sampler.p.backup")



import numpy
from . import center
from .libinterferometry import grid
from ..imaging import Image
from ..constants.astronomy import arcsec
from scipy.fftpack import ifft2, fftshift, ifftshift, fftfreq

def invert(data, imsize=256, pixel_size=0.25, convolution="pillbox", mfs=False,\
        weighting="natural", robust=2, npixels=0, centering=None, \
        mode='continuum', beam=False, uvtaper=None):

    # If we are calculating the beam, set all of the real values to 1 and the
    # imaginary data to 0.

    if beam:
        real = data.real.copy()
        imag = data.imag.copy()

        data.real[:,:] = 1.
        data.imag[:,:] = 0.

    # Apply a uv-taper to the data, if requested.

    if type(uvtaper) != type(None):
        taper = numpy.exp(-0.5*data.uvdist**2/(uvtaper*1e3)**2)

        weights = data.weights.copy()
        for i in range(data.freq.size):
            data.weights[:,i] *= taper

    # Grid the data before imaging.
    
    binsize = 1.0 / (pixel_size * imsize * arcsec)
    gridded_data = grid(data, gridsize=imsize, binsize=binsize, \
            convolution=convolution, mfs=mfs, imaging=True, \
            weighting=weighting, robust=robust, npixels=npixels, mode=mode)

    # If making an image of the beam, restore the data to what it was before.

    if beam:
        data.real = real
        data.imag = imag

    # If using uvtaper, restore the weights.

    if type(uvtaper) != type(None):
        data.weights = weights

    # Center the data, if requested.

    if type(centering) != type(None):
        gridded_data = center(gridded_data, centering)

    # Set up information for the final image.
            
    x = fftshift(fftfreq(imsize, binsize)) / arcsec
    y = fftshift(fftfreq(imsize, binsize)) / arcsec
    xx, yy = numpy.meshgrid(x, y)
    r = numpy.sqrt(xx**2 + yy**2)

    image = numpy.zeros((imsize,imsize,gridded_data.real.shape[1],1))

    # Loop through frequency space and do the Fourier transform.

    for i in range(gridded_data.real.shape[1]):
        u = gridded_data.u.reshape((imsize, imsize))
        v = gridded_data.v.reshape((imsize, imsize))
        real = gridded_data.real[:,i].reshape((imsize, imsize))
        imag = gridded_data.imag[:,i].reshape((imsize, imsize))
        weights = gridded_data.weights[:,i].reshape((imsize, imsize))

        comp = real + 1j*imag
        
        if convolution == "pillbox":
            conv_func = pillbox
        elif convolution == "expsinc":
            conv_func = exp_sinc

        im = fftshift(ifft2(ifftshift(comp))).real * imsize**2

        convolve = fftshift(ifft2(ifftshift(conv_func(u, v, binsize, \
               binsize)))).real

        image[:,:,i,0] = (im/convolve)[:,::-1]

        # Make sure the beam peaks at exactly 1. Should be just a small 
        # correction.

        if beam:
            image[:,:,i,0] /= image[:,:,i,0].max()

    return Image(image, x=x, y=y, freq=gridded_data.freq)

def pillbox(u, v, delta_u, delta_v):

    m = 1
    
    arr = numpy.ones(u.shape, dtype=float)*u.size

    arr[numpy.abs(u) >= m * delta_u / 2] = 0
    arr[numpy.abs(v) >= m * delta_v / 2] = 0

    return arr

def exp_sinc(u, v, delta_u, delta_v):
    
    alpha1 = 1.55
    alpha2 = 2.52
    m = 6
    
    arr = numpy.sinc(u / (alpha1 * delta_u)) * \
            numpy.exp(-1 * (u / (alpha2 * delta_u))**2)* \
            numpy.sinc(v / (alpha1 * delta_v)) * \
            numpy.exp(-1 * (v / (alpha2 * delta_v))**2)

    arr[numpy.abs(u) >= m * delta_u / 2] = 0
    arr[numpy.abs(v) >= m * delta_v / 2] = 0

    arr = arr/arr.sum() * arr.size

    return arr



from .readuvfits import readuvfits
from .libinterferometry import Visibilities
from glob import glob
import numpy

def readvis(filename, fmt="casa"):

    filenames = numpy.array(glob(filename+"/*.uv.fits"))

    for i in range(filenames.size):
        if i == 0:
            vis = readuvfits(filenames[i],fmt=fmt)

            vis.u /= vis.freq.mean()
            vis.v /= vis.freq.mean()
        else:
            new = readuvfits(filenames[i],fmt=fmt)

            vis.freq = numpy.concatenate((vis.freq,new.freq))
            vis.real = numpy.concatenate((vis.real,new.real),axis=1)
            vis.imag = numpy.concatenate((vis.imag,new.imag),axis=1)
            vis.amp = numpy.concatenate((vis.amp,new.amp),axis=1)
            vis.weights = numpy.concatenate((vis.weights,new.weights),axis=1)

    vis.real = vis.real[:,numpy.argsort(vis.freq)][:,::-1]
    vis.imag = vis.imag[:,numpy.argsort(vis.freq)][:,::-1]
    vis.amp = vis.amp[:,numpy.argsort(vis.freq)][:,::-1]
    vis.weights = vis.weights[:,numpy.argsort(vis.freq)][:,::-1]
    vis.freq = vis.freq[numpy.argsort(vis.freq)][::-1]

    vis.u *= vis.freq.mean()
    vis.v *= vis.freq.mean()

    return Visibilities(vis.u, vis.v, vis.freq, vis.real, vis.imag, \
            vis.weights, baseline=vis.baseline)



from numpy import sqrt, exp, pi

def ml2d(x, y, z, sigma_z, params, model, args=None, limits=None):

    if (type(limits) != type(None)):
        for i in range(params.size):
            if ((limits[i]["limited"][0] == True) and \
                    (params[i] < limits[i]["limits"][0])):
                ml = 0.0
                chisq = 1.0e300
                return ml, chisq
            elif ((limits[i]["limited"][1] == True) and \
                    (params[i] > limits[i]["limits"][1])):
                ml = 0.0
                chisq = 1.0e300
                return ml, chisq

    m = model(x, y, params, **args)

    mlarr = 1./sqrt(2*pi*sigma_z**2)*\
                exp(-(z-m)**2/(2*sigma_z**2)) 

    ml = (mlarr**(1./(z.size-params.size))).prod()
    chisq = ((z-m)**2/sigma_z**2).sum()

    return ml, chisq



from pdspy.constants.astronomy import R_sun, pc
import scipy.interpolate
import astropy.table
import numpy

def btsettl_photometry(Teff=5770., Logg=4.4, Rstar=1.0, dpc=140., \
        system="Johnson", filters="B", AB=False):

    # Create the filename.

    path = '/'.join(__file__.split("/")[0:-1])+"/btsettl_data/"

    basename = "colmag.BT-Settl.server."

    filename = basename + system + "."

    if AB:
        filename += "AB.txt"
    else:
        filename += "Vega.txt"

    # Get the column names.

    f = open(path+filename, "r")
    lines = f.readlines()
    f.close()

    for i, line in enumerate(lines):
        if line[0] != "!":
            break

    colnames = lines[i-1].split()[1:]

    # Get the data.

    data = numpy.loadtxt(path+filename, comments='!')

    # Make the data into a table.

    table = astropy.table.Table(data[1:,0:16], names=colnames)

    # Make sure we only include the Teff's with all the log(g)'s.

    """
    teff = numpy.unique(table["Teff"])
    logg = numpy.unique(table["Logg"])

    for T in teff:
        if len(table["Teff"][table["Teff"] == T]) < len(logg):
            table.remove_rows(numpy.where(table["Teff"] == T)[0])
    """

    # Now loop through the filters and get the photometry.

    """
    teff = numpy.array(numpy.unique(table["Teff"]))
    logg = numpy.array(numpy.unique(table["Logg"]))
    """

    mag = []

    for filter in filters.split(','):
        #data = numpy.array(table[filter]).reshape((teff.size, logg.size))

        # Now do the 2D interpolation.

        #f = scipy.interpolate.interp2d(logg, teff, data)

        f = scipy.interpolate.LinearNDInterpolator((table["Teff"], \
                table["Logg"]), table[filter], rescale=False)

        # Calculate the photometry now.

        mag.append(f([[Teff,Logg]])[0])

    # Turn the photometry into an array.

    mag = numpy.array(mag)

    # Turn into an absolute magnitude.

    M = mag - 5 * numpy.log10(Rstar*R_sun / pc) + 5.

    # Now get the apparent magnitude.

    photometry = M + 5*(numpy.log10(dpc) - 1.)

    return photometry



import numpy

def ttest(data1, data2, tailed=2):

    n = len(data1)
    m = len(data2)
    nu = n + m - 2

    mean1 = numpy.mean(data1)
    mean2 = numpy.mean(data2)

    S1 = numpy.sum((data1 - mean1)**2) / n
    S2 = numpy.sum((data2 - mean2)**2) / m

    s = (n*S1 + m*S2) / nu

    t = (mean1 - mean2) / (numpy.sqrt(s) * numpy.sqrt(m**-1 + n**-1))

    T = numpy.random.standard_t(nu, size=100000)

    if tailed == 1:
        if t < 0:
            p = len(T[T < t]) / len(T)
        elif t > 0:
            p = len(T[T > t]) / len(T)
    elif tailed == 2:
        if t < 0:
            p = len(T[(T < t) | (T > -t)]) / len(T)
        elif t > 0:
            p = len(T[(T > t) | (T < -t)]) / len(T)

    return t, p



import astropy.table

class MaskedColumn(astropy.table.MaskedColumn):

    def __getitem__(self, item):
        x = super(MaskedColumn, self).__getitem__(item)

        try:
            if x.mask:
                return '--'
        except:
            return x



from setuptools import setup
from numpy.distutils.core import setup, Extension
from Cython.Build import cythonize

from distutils.command.sdist import sdist
cmdclass={'sdist': sdist}

# Set up the extension modules.

libinterferometry = cythonize([\
        Extension('pdspy.interferometry.libinterferometry',\
            ["pdspy/interferometry/libinterferometry.pyx"],\
            libraries=["m"], extra_compile_args=['-ffast-math'])])[0]

libimaging = cythonize([Extension('pdspy.imaging.libimaging',\
        ["pdspy/imaging/libimaging.pyx"], libraries=[], \
        extra_compile_args=[])])[0]

bhmie = Extension('pdspy.dust.bhmie', sources=['pdspy/dust/bhmie.f90'])

bhcoat = Extension('pdspy.dust.bhcoat', sources=['pdspy/dust/bhcoat.f90'])

dmilay = Extension('pdspy.dust.dmilay', sources=['pdspy/dust/DMiLay.f90'])

read = cythonize([Extension('pdspy.radmc3d.read', ["pdspy/radmc3d/read.pyx"], \
        libraries=[], extra_compile_args=[])])[0]

# Now define the setup for the package.

setup(name="pdspy", \
        version="2.0.6", \
        author="Patrick Sheehan", \
        author_email="psheehan@northwestern.edu", \
        description="Radiative transfer modeling of protoplanetary disks", \
        long_description=open("README.md","r").read(), \
        long_description_content_type="text/markdown", \
        url="https://github.com/psheehan/pdspy", \
        packages=[\
        "pdspy",\
        "pdspy.constants", \
        "pdspy.dust",\
        "pdspy.gas",\
        "pdspy.imaging",\
        "pdspy.interferometry", \
        "pdspy.mcmc",\
        "pdspy.misc",\
        "pdspy.modeling",\
        "pdspy.plotting", \
        "pdspy.radmc3d",\
        "pdspy.spectroscopy",\
        "pdspy.stars",\
        "pdspy.statistics", \
        "pdspy.table", \
        "pdspy.utils"], \
        package_dir={\
        "pdspy.dust": 'pdspy/dust', \
        "pdspy.gas": 'pdspy/gas', \
        "pdspy.spectroscopy": 'pdspy/spectroscopy', \
        "pdspy.stars": 'pdspy/stars'}, \
        package_data={\
        'pdspy.dust': ['data/*','reddening/*.dat'], \
        'pdspy.imaging': ['*.pyx'], \
        'pdspy.interferometry': ['*.pyx'], \
        'pdspy.gas': ['data/*.dat'], \
        'pdspy.radmc3d': ['*.pyx'], \
        'pdspy.spectroscopy': ['btsettle_data/*.txt']}, \
        ext_modules=[libinterferometry, libimaging, bhmie, \
        bhcoat, dmilay, read], \
        scripts=[\
        'bin/config_template.py',\
        'bin/upgrade_to_pdspy2.py',\
        'bin/generate_surrogate_model.py',\
        'bin/disk_model_emcee3.py',\
        'bin/disk_model_nested.py',\
        'bin/disk_model_dynesty.py',\
        'bin/disk_model_powerlaw.py',\
        'bin/flared_model_emcee3.py',\
        'bin/flared_model_nested.py', \
        'bin/flared_model_dynesty.py'], \
        install_requires=['numpy','scipy','matplotlib','emcee','corner',\
        'hyperion','h5py','mpi4py','Cython','astropy','schwimmbad','dynesty',\
        'scikit-learn'], \
        cmdclass=cmdclass)




#!/usr/bin/env python3

from pdspy.dust import Dust, DustGenerator
import numpy
import matplotlib
import matplotlib.pyplot as plt

# Maximum dust grain sizes.

a_max = numpy.logspace(0.,5.,1000)

# Grain size distribution power-law

p = numpy.linspace(2.5, 4.5, 10)

# Read in the dust generator class.

dust_gen = DustGenerator("diana_wice.hdf5")

# Change a few of the parameters to make the plot look nice.

matplotlib.rcParams["font.family"] = "serif"
matplotlib.rcParams["font.size"] = 14
matplotlib.rcParams["text.usetex"] = "True"
matplotlib.rcParams["text.latex.preamble"] = r"\usepackage{upgreek}"
matplotlib.rcParams["legend.fontsize"] = 14

# Start the plotting.

fig, ax = plt.subplots(nrows=1, ncols=1)

# Loop through each of the values for p.

for i in range(p.size):
    # Set up a list to put beta in.

    kabs = []

    # Loop through and calculate beta.

    for j in range(a_max.size):
        # Get the dust generator properties.

        dust = dust_gen(a_max[j] / 1.0e4, p[i])

        # Add to the array.

        kabs.append(dust.kabs[103])

    # Plot beta.

    ax.semilogx(a_max, kabs, "-", label="$p = {0:3.1f}$".format(p[i]))

# Add a legend.

plt.legend(loc="lower left")

# Adjust the figure.

ax.set_xlabel("$a_{max}$ [$\mu$m]")
ax.set_ylabel(r"$\kappa_{abs}$")

# Save the figure.

fig.savefig("kabs_plot.pdf")



from .load_config import load_config
from .load_data import load_data
from .load_results import load_results
from .propose_point_emcee import propose_point_emcee

from . import emcee
from . import dynesty



#!/usr/bin/env python3

from pdspy.constants.astronomy import arcsec
import pdspy.interferometry as uv
import pdspy.spectroscopy as sp
import pdspy.imaging as im
import pdspy.modeling as modeling
import pdspy.plotting as plotting
import pdspy.utils as utils
import matplotlib.pyplot as plt
import astropy.stats
import schwimmbad
import argparse
import numpy
import sys
import os
import emcee
import corner
from mpi4py import MPI

comm = MPI.COMM_WORLD

################################################################################
#
# Parse command line arguments.
#
################################################################################

parser = argparse.ArgumentParser()
parser.add_argument('-o', '--object')
parser.add_argument('-r', '--resume', action='store_true')
parser.add_argument('-c', '--withhyperion', action='store_true')
parser.add_argument('-p', '--resetprob', action='store_true')
parser.add_argument('-a', '--action', type=str, default="run")
parser.add_argument('-n', '--ncpus', type=int, default=1)
parser.add_argument('-m', '--ncpus_highmass', type=int, default=8)
parser.add_argument('-e', '--withexptaper', action='store_true')
parser.add_argument('-t', '--timelimit', type=int, default=7200)
parser.add_argument('-b', '--trim', type=str, default="")
parser.add_argument('-s', '--SED', action='store_true')
parser.add_argument('-i', '--nice', action='store_true')
parser.add_argument('-l', '--nicelevel', type=int, default=19)
parser.add_argument('-f', '--ftcode', type=str, default="galario")
args = parser.parse_args()

# Check whether we are using MPI.

withmpi = comm.Get_size() > 1

# Set the number of cpus to use.

ncpus = args.ncpus
ncpus_highmass = args.ncpus_highmass

# Set the nice level of the subprocesses.

if args.nice:
    nice = args.nicelevel
else:
    nice = None

# Get the source name and check that it has been set correctly.

source = args.object

if source == None:
    print("--object must be specified")
    sys.exit()

# Determine what action should be taken and set some variables appropriately.

if args.action not in ['run','plot']:
    print("Please select a valid action")
    sys.exit(0)

if args.action == 'plot':
    args.resume = True

# Get the items we want to trim.

if args.trim == "":
    trim = []
else:
    trim = args.trim.split(",")

################################################################################
#
# Create a function which returns a model of the data.
#
################################################################################

# Define a likelihood function.

def lnlike(params, visibilities, images, spectra, parameters, plot):

    m = modeling.run_disk_model(visibilities, images, spectra, params, \
            parameters, plot, ncpus=ncpus, ncpus_highmass=ncpus_highmass, \
            with_hyperion=args.withhyperion, timelimit=args.timelimit, \
            source=source, nice=nice, run_thermal=False)

    # Catch whether the model timed out.

    if m == 0.:
        return -numpy.inf

    # A list to put all of the chisq into.

    chisq = []

    # Calculate the chisq for the visibilities.

    for j in range(len(visibilities["file"])):
        chisq.append(-0.5*(numpy.sum((visibilities["data"][j].real - \
                m.visibilities[visibilities["lam"][j]].real)**2 * \
                visibilities["data"][j].weights)) + \
                -0.5*(numpy.sum((visibilities["data"][j].imag - \
                m.visibilities[visibilities["lam"][j]].imag)**2 * \
                visibilities["data"][j].weights)))

    # Calculate the chisq for all of the images.

    for j in range(len(images["file"])):
        chisq.append(-0.5 * (numpy.sum((images["data"][j].image - \
                m.images[images["lam"][j]].image)**2 / \
                images["data"][j].unc**2)))

    # Calculate the chisq for the SED.

    if "total" in spectra:
        chisq.append(-0.5 * (numpy.sum((spectra["total"].flux - \
                m.spectra["SED"].flux)**2 / spectra["total"].unc**2)))

    # Return the sum of the chisq.

    return numpy.array(chisq).sum()

# Define a prior function.

def lnprior(params, parameters, visibilities):
    for key in parameters:
        if not parameters[key]["fixed"]:
            if parameters[key]["limits"][0] <= params[key] <= \
                    parameters[key]["limits"][1]:
                pass
            else:
                return -numpy.inf

    # Make sure that the radii are correct.

    if "logR_in" in params:
        R_in = 10.**params["logR_in"]
    else:
        R_in = 10.**parameters["logR_in"]["value"]

    if "logR_disk" in params:
        R_disk = 10.**params["logR_disk"]
    else:
        R_disk = 10.**parameters["logR_disk"]["value"]

    if "logR_env" in params:
        R_env = 10.**params["logR_env"]
    else:
        R_env = 10.**parameters["logR_env"]["value"]

    if R_in <= R_disk <= R_env:
        pass
    else:
        return -numpy.inf

    # Check that we aren't allowing any absurdly dense models.

    if "logR_env" in params and "logM_env" in params:
        if params["logR_env"] < 0.5 * params["logM_env"] + 4.:
            return -numpy.inf
        else:
            pass

    # Check that the cavity actually falls within the disk.

    if not parameters["logR_cav"]["fixed"]:
        if R_in <= 10.**params["logR_cav"] <= R_disk:
            pass
        else:
            return -numpy.inf

    # Check that the gap is reasonable.

    if not parameters["logR_gap1"]["fixed"]:
        if R_in <= 10.**params["logR_gap1"] - params["w_gap1"]/2:
            pass
        else:
            return -numpy.inf

    # Everything was correct, so continue on.

    lnp = 0.0

    # Add in the priors.

    for i in range(len(visibilities["file"])):
        if (not parameters["flux_unc{0:d}".format(i+1)]["fixed"]) and \
                (parameters["flux_unc{0:d}".format(i+1)]["prior"] == "gaussian"):
            lnp += -0.5 * (params["flux_unc{0:d}".format(i+1)] - \
                    parameters["flux_unc{0:d}".format(i+1)]["value"])**2 / \
                    parameters["flux_unc{0:d}".format(i+1)]["sigma"]**2

    # Return

    return lnp

# Define a probability function.

def lnprob(p, visibilities, images, spectra, parameters, plot):

    keys = []
    for key in sorted(parameters.keys()):
        if not parameters[key]["fixed"]:
            keys.append(key)

    params = dict(zip(keys, p))

    lp = lnprior(params, parameters, visibilities)

    if not numpy.isfinite(lp):
        return -numpy.inf

    return lp + lnlike(params, visibilities, images, spectra, parameters, \
            plot)

################################################################################
#
# In case we are restarting this from the same job submission, delete any
# temporary directories associated with this run.
#
################################################################################

os.system("rm -r /tmp/temp_{0:s}_{1:d}".format(source, comm.Get_rank()))

################################################################################
#
# Set up a pool for parallel runs.
#
################################################################################

if args.action == "run":
    if withmpi:
        pool = schwimmbad.MPIPool()

        if not pool.is_master():
            pool.wait()
            sys.exit(0)
    else:
        pool = None

################################################################################
#
# Read in the data.
#
################################################################################

# Import the configuration file information.

config = utils.load_config()

# Deprecated option to include an expontial taper on the disk.

if args.withexptaper:
    config.parameters["disk_type"]["value"] = "exptaper"

# Read in the data.

visibilities, images, spectra = utils.load_data(config)

################################################################################
#
# Fit the model to the data.
#
################################################################################

# Set up the emcee run.

ndim = 0
keys = []

for key in sorted(config.parameters.keys()):
    if not config.parameters[key]["fixed"]:
        ndim += 1
        keys.append(key)

# Make the labels nice with LaTeX.

labels = ["$"+key.replace("_","_{").replace("log","\log ")+"}$" \
        if key[0:3] == "log" else "$"+key.replace("h_large","h,large")+\
        "$" for key in keys]

# If we are resuming an MCMC simulation, read in the necessary info, otherwise
# set up the info.

if args.resume:
    pos = numpy.load("pos.npy")
    chain = numpy.load("chain.npy")
    state = None
    nsteps = chain[0,:,0].size

    if args.resetprob:
        prob = None
        prob_list = numpy.empty((config.nwalkers,0))
    else:
        prob_list = numpy.load("prob.npy")
        if len(prob_list.shape) == 1:
            prob_list = prob_list.reshape((config.nwalkers,1))
        prob = prob_list[:,-1]
else:
    pos = []
    for j in range(config.nwalkers):
        pos.append(utils.propose_point_emcee(config.parameters))

    prob = None
    prob_list = numpy.empty((config.nwalkers, 0))
    chain = numpy.empty((config.nwalkers, 0, ndim))
    state = None
    nsteps = 0

# Set up the MCMC simulation.

if args.action == "run":
    sampler = emcee.EnsembleSampler(config.nwalkers, ndim, lnprob, \
            args=(visibilities, images, spectra, config.parameters, False), \
            pool=pool)

# Run a few burner steps.

if args.action == "plot":
    nsteps = config.max_nsteps-1

while nsteps < config.max_nsteps:
    if args.action == "run":
        for i in range(config.steps_per_iter):
            pos, prob, state = sampler.run_mcmc(pos, 1, lnprob0=prob, \
                    rstate0=state)

            chain = numpy.concatenate((chain, sampler.chain), axis=1)

            prob_list = numpy.concatenate((prob_list, prob.\
                    reshape((config.nwalkers,1))), axis=1)

            # Plot the steps of the walkers.

            for j in range(ndim):
                fig, ax = plt.subplots(nrows=1, ncols=1)

                for k in range(config.nwalkers):
                    ax.plot(chain[k,:,j])

                plt.savefig("steps_{0:s}.png".format(keys[j]))

                plt.close(fig)

            # Save walker positions in case the code stps running for some 
            # reason.

            numpy.save("pos", pos)
            numpy.save("prob", prob_list)
            numpy.save("chain", chain)

            # Augment the nuber of steps and reset the sampler for the next run.

            nsteps += 1

            sampler.reset()

    # Get the best fit parameters and uncertainties from the last 10 steps.

    samples = chain[:,-config.nplot:,:].reshape((-1, ndim))

    # Make the cuts specified by the user.

    for command in trim:
        command = command.split(" ")

        for i, key in enumerate(keys):
            if key == command[0]:
                if command[1] == '<':
                    good = samples[:,i] > float(command[2])
                else:
                    good = samples[:,i] < float(command[2])

                samples = samples[good,:]

    # Get the best fit parameters.

    params = numpy.median(samples, axis=0)
    sigma = astropy.stats.mad_std(samples, axis=0)

    # Write out the results.

    f = open("fit.txt", "w")
    f.write("Best fit parameters:\n\n")
    for j in range(len(keys)):
        f.write("{0:s} = {1:f} +/- {2:f}\n".format(keys[j], params[j], \
                sigma[j]))
    f.write("\n")
    f.close()

    os.system("cat fit.txt")

    # Plot histograms of the resulting parameters.

    fig = corner.corner(samples, labels=labels, truths=params)

    plt.savefig("fit.pdf")

    # Make a dictionary of the best fit parameters.

    params = dict(zip(keys, params))

    ############################################################################
    #
    # Plot the results.
    #
    ############################################################################

    # Plot the best fit model over the data.

    fig, ax = plt.subplots(nrows=2*len(visibilities["file"]), ncols=3)

    # Create a high resolution model for averaging.

    m = modeling.run_disk_model(visibilities, images, spectra, params, \
            config.parameters, plot=True, ncpus=ncpus, \
            ncpus_highmass=ncpus_highmass, with_hyperion=args.withhyperion, \
            timelimit=args.timelimit, source=source, nice=nice, \
            run_thermal=False, ftcode=args.ftcode)

    # Plot the millimeter data/models.

    for j in range(len(visibilities["file"])):
        # Plot the visibilities.

        plotting.plot_1D_visibilities(visibilities, m, config.parameters, \
                params, index=j, fig=(fig, ax[2*j,0]))

        # Plot the 2D visibilities.

        plotting.plot_2D_visibilities(visibilities, m, config.parameters, \
                params, index=j, fig=(fig, ax[2*j+1,0:2]))

        # Plot the model image.

        plotting.plot_continuum_image(visibilities, m, config.parameters, \
                params, index=j, fig=(fig, ax[2*j,1]))

    # Plot the SED.

    plotting.plot_SED(spectra, m, SED=args.SED, fig=(fig, ax[0,2]))

    # Plot the scattered light image.

    for j in range(len(images["file"])):
        plotting.plot_scattered_light(visibilities, m, config.parameters, \
                params, index=j, fig=(fig, ax[1,2]))

    # Turn off axes when they aren't being used.

    if len(images["file"]) == 0:
        ax[1,2].set_axis_off()

    for j in range(len(visibilities["file"])):
        if j > 0:
            ax[2*j,2].set_axis_off()
            ax[2*j+1,2].set_axis_off()

    # Adjust the plot.

    fig.set_size_inches((12.5,8*len(visibilities["file"])))
    fig.subplots_adjust(left=0.07, right=0.98, top=0.95, bottom=0.08, \
            wspace=0.25, hspace=0.2)

    # Save the figure.

    fig.savefig("model.pdf")

    plt.close(fig)

    if args.action == "plot":
        nsteps = numpy.inf

# Now we can close the pool and end the code.

if args.action == "run":
    if withmpi:
        pool.close()



class NumberUnit:

    def __init__(self,cgs):
        self.cgs = cgs

    def __add__(self,a):
        return self.cgs + a

    def __radd__(self,a):
        return self.cgs + a

    def __sub__(self,a):
        return self.cgs - a

    def __rsub__(self,a):
        return a - self.cgs

    def __mul__(self,a):
        return self.cgs * a

    def __rmul__(self,a):
        return self.cgs * a

    def __div__(self,a):
        return self.cgs / a

    def __rdiv__(self,a):
        return a / self.cgs



#!/usr/bin/env python3

import argparse
import numpy
import os
import pyDOE

import pdspy.modeling as modeling
import pdspy.utils as utils

from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
import pickle
import glob

import torch
import gpytorch

from torch.utils.data import TensorDataset, DataLoader
from gpytorch.models import ApproximateGP
from gpytorch.variational import CholeskyVariationalDistribution
from gpytorch.variational import VariationalStrategy, \
        IndependentMultitaskVariationalStrategy

import astropy.stats

################################################################################
#
# Command line arguments.
#
################################################################################

parser = argparse.ArgumentParser()
parser.add_argument('-i', '--index', type=int, default=0)
parser.add_argument('-t', '--total_array_tasks', type=int, default=1)
parser.add_argument('-a', '--action')
parser.add_argument('-n', '--ncpus', type=int, default=1)
parser.add_argument('-l', '--nlearn', type=int, default=1)
parser.add_argument('-g', '--grid', type=str, default="train")
args = parser.parse_args()

################################################################################
#
# Set up the parameters of the grid.
#
################################################################################

config = utils.load_config()

keys = []
for key in config.parameters:
    if not config.parameters[key]["fixed"]:
        keys.append(key)

################################################################################
#
# Create a function to create a set of samples of parameter space.
#
################################################################################

def generate_samples(nsamples, mode="LHS", grid_keys=keys):
    # Create the grid of parameters for models using a latin hypercube sampling:

    if mode == "LHS":
        samples = pyDOE.lhs(len(grid_keys), samples=nsamples, \
                criterion="maximin")
    else:
        samples = numpy.random.uniform(size=(nsamples,len(keys)))

    # Stretch the samples to the appropriate range.

    for i, key in enumerate(grid_keys):
        #limits = grid_params[key+"_range"]
        limits = config.parameters[key]["limits"]
        samples[:,i] = (limits[1] - limits[0])*samples[:,i] + limits[0]

    # Now turn this into an array that we can save.

    pars = []

    for i in range(nsamples):
        #pars.append(dict(zip(keys, samples[i,:])))

        pars.append(dict(zip(keys, [samples[i,\
                numpy.where(numpy.array(grid_keys) == par)[0][0]] if par in \
                grid_keys else config.parameters[par]["value"] for par in \
                keys])))

        # Now get the filename.

        pars[-1]["filename"] = "".join([key.replace("_","")+"_"+ \
                "{0:.5f}".format(pars[-1][key]).replace(".hdf5","").\
                #replace("pollack_", "").replace("88","").replace("66","")+ \
                replace("pollack_", "")+ \
                "_" for key in keys]).rstrip("_")+".hdf5"

    print(len(pars))
    parameters = numpy.array(pars)

    return parameters, samples

################################################################################
#
# Convenience function for getting the proper model directory.
#
################################################################################

def get_directory(grid="train", key=None):
    directory = grid.capitalize().replace("d","D") + "Grid" + \
            ("s" if "1" in grid else "") + "/"

    if key is not None:
        directory = directory + key + "/"

    return directory

################################################################################
#
# Create a function to run models from parameters.npy
#
################################################################################

def run_models(grid="train", key=None):
    # Get the directory to load from.

    directory = get_directory(grid=grid, key=key)

    # Load the parameters file.

    parameters = numpy.load(directory+"parameters.npy", allow_pickle=True)
    print(parameters.size)

    # Load the list of models that have been tried.

    tried_models = numpy.loadtxt(directory+"tried_models.txt", dtype=str)

    # Run the model:

    start = args.index
    step = args.total_array_tasks

    bad = []

    for i in range(start,parameters.size,step):
        base_dir = os.environ["PWD"]

        print(parameters[i])

        if i >= len(parameters):
            continue
        elif os.path.exists(directory+parameters[i]["filename"]):
            continue
        elif parameters[i]["filename"] in tried_models:
            continue

        if not os.path.exists(directory):
            os.mkdir(directory)

        try:
            model = modeling.run_disk_model(config.visibilities, config.images,\
                    config.spectra, parameters[i], config.parameters, \
                    timelimit=7200, with_hyperion=True, \
                    percentile=config.percentile, absolute=config.absolute, \
                    relative=config.relative, ncpus=args.ncpus, \
                    ncpus_highmass=args.ncpus, verbose=True, \
                    increase_photons_until_convergence=True)

            # Write out the file.

            model.write_yso(directory+parameters[i]["filename"])
        except:
            f = open(directory+"tried_models.txt","a")
            f.write(parameters[i]["filename"]+"\n")
            f.close()
            #bad.append(i)

        """
        new_parameters = numpy.delete(parameters, bad)
        numpy.save(directory+"parameters.npy", new_parameters)
        """

################################################################################
#
# Load in the model grid.
#
################################################################################

def load_data(grid="train", key=None, pca=False):
    # Get the directory to load from.

    directory = get_directory(grid=grid, key=key)

    # Load in the parameters file with all of the model info.

    parameters = numpy.load(directory+"parameters.npy", allow_pickle=True)

    # Check whether a data file already exists.

    if os.path.exists(directory+"data.npz"):
        x = numpy.load(directory+"data.npz")['x'].tolist()
        data = numpy.load(directory+"data.npz")['data'].tolist()
        original = numpy.load(directory+"data.npz")['original'].tolist()
        filenames = numpy.load(directory+"data.npz")['filenames'].tolist()
        shape = tuple(numpy.load(directory+"data.npz")['shape'])

        start_index = len(data)
    else:
        data = []
        original = []
        x = []
        filenames = []

        start_index = 0

    if len(glob.glob(directory+"*.hdf5")) > len(data):
        for p in parameters[start_index:]:
            try:
                m = modeling.YSOModel()
                m.read_yso(directory+p["filename"])

                temperature = numpy.array(m.grid.temperature)
                density = numpy.array(m.grid.density)

                density[density == 0] = 1.0e-30

                original.append(temperature)

                data.append(((temperature[:,:,:,0]*density[:,:,:,0]).\
                        sum(axis=0) / density[:,:,:,0].sum(axis=0)).flatten())

                x.append([p[key] for key in keys])
                filenames.append(p["filename"])
            except:
                pass

    x = numpy.array(x)
    data = numpy.array(data)
    original = numpy.array(original)
    #shape = (99, 100, 1)
    try:
        shape = temperature[0].shape
    except:
        pass

    numpy.savez(directory+"data.npz", data=data, x=x, filenames=filenames, \
            original=original, shape=shape)

    data[data <= 0] = 0.1
    data = numpy.log10(data)

    if pca:
        # Load in the PCA that was found.

        pca = pickle.load(open("pca.pkl","rb"))

        # Load in the data.

        y = pca.transform(data)
        ncomponents = y.shape[1]

        return x, y, ncomponents
    else:
        return x, data, shape

################################################################################
#
# Create a function to do the PCA decomposition.
#
################################################################################

def run_pca_decomposition(plot=False):
    # First, load in the data.

    x, data, shape = load_data(grid="train", pca=False)
    print(data.shape)

    # Make some plots, if requested.

    if plot:
        # Compute a PCA decomposition.

        pca = PCA(n_components=100)
        pca.fit(data)

        # Plot the variance explained.

        fig, ax = plt.subplots(nrows=1, ncols=1, gridspec_kw=dict(left=0.2, \
                right=0.96, top=0.95, bottom=0.15))

        ax.plot(numpy.cumsum(pca.explained_variance_ratio_)*100)

        ax.set_xlabel("Number of PCA Components", fontsize=14, labelpad=10)
        ax.set_ylabel("Cumulative Variance Explained (\%)", fontsize=14, \
                labelpad=10)

        ax.tick_params(labelsize=14)

        ax.set_xlim(0,100)
        ax.set_ylim(80.,100.)

        #plt.show()
        plt.savefig("variance.png")

        # Get the eigentempertemperatures.

        fig, axes = plt.subplots(3, 5, figsize=(12,7), subplot_kw={"xticks":[],\
                "yticks":[], 'projection':'polar'}, gridspec_kw=dict(left=0.03,\
                right=0.97, top=0.95, bottom=0.02, hspace=0.3, wspace=0.15))

        r, theta = numpy.meshgrid(numpy.arange(shape[0]), numpy.linspace(0., \
                numpy.pi/2, shape[1])[::-1], indexing='ij')

        for i, ax in enumerate(axes.flat):
            ax.pcolor(theta, r, pca.components_[i].reshape(shape)[:,:,0], \
                    shading="auto")

            ax.set_title("Eigentemperature, $T_{{{0:d}}}$".format((i+1)))

            ax.set_thetalim(0.,numpy.pi/2)

            ax.set_rticks([])
            ax.set_xticks([])

        plt.savefig("eigentemperatures.png")

        # Now plot the reconstrcuted temperature for the first 9 
        # eigentemperatures.

        nrows = 16

        fig, ax = plt.subplots(2, nrows, figsize=(17.5,2.5), gridspec_kw=dict(\
                hspace=0.02,wspace=0.1, left=0.03, right=0.99, bottom=0.015, \
                top=0.92),\
                subplot_kw={'xticks':[], 'yticks':[], 'projection':'polar'})

        index = numpy.random.randint(data.shape[0])
        print("Plotting reconstruction for parameters:")
        print(dict(zip(keys, x[index,:])))

        for i in range(nrows):
            if i == nrows-1:
                ax[0,i].pcolor(theta, r, 10.**data[index].reshape(shape)\
                        [:,:,0], shading="auto")
                ax[1,i].set_axis_off()

                ax[0,i].set_rticks([])
                ax[0,i].set_xticks([])
                ax[1,i].set_rticks([])
                ax[1,i].set_xticks([])

                ax[0,i].set_thetalim(0., numpy.pi/2)

                ax[0,i].set_title("Actual")
            else:
                pca = PCA(n_components=int(3*i+1)).fit(data)
                components = pca.transform(data[index+0:index+1,:])
                projected = pca.inverse_transform(components)

                ax[0,i].pcolor(theta, r, 10.**projected[0].reshape(shape)\
                        [:,:,0], vmin=10.**data[index].min(), \
                        vmax=10.**data[index].max(), shading="auto")

                diff = 10.**projected[0] - 10.**data[index]

                ax[1,i].pcolor(theta, r, diff.reshape(shape)[:,:,0], \
                        vmin=-numpy.abs(diff).max(), vmax=numpy.abs(diff).\
                        max(), shading="auto")

                ax[0,i].set_title("N = {0:d}".format(int(3*i+1)))

                print("Maximum Error for {1:d} Components = {0:4.2f} K".format(\
                        numpy.abs(diff).max(), int(3*i+1)))

                ax[0,i].set_rticks([])
                ax[0,i].set_xticks([])
                ax[1,i].set_rticks([])
                ax[1,i].set_xticks([])

                ax[0,i].set_thetalim(0., numpy.pi/2)
                ax[1,i].set_thetalim(0., numpy.pi/2)

        ax[0,0].set_ylabel("Reconstructed\nTemperature")
        ax[1,0].set_ylabel("Reconstructed\n- Actual")

        #plt.show()
        plt.savefig("pca_reconstruction.png")

    # Now that we know how many features to use, generate the final PCA and 
    # save it.

    pca = PCA(n_components=50).fit(data)

    pickle.dump(pca, (open("pca.pkl","wb")))

################################################################################
#
# Create a Gaussian Process model to be used.
#
################################################################################

class GPModel(ApproximateGP):
    def __init__(self, inducing_points=None, ncomponents=1, load=False):
        if load:
            inducing_points = torch.load("inducing_points.pt")

            pca = pickle.load(open("pca.pkl","rb"))
            ncomponents = pca.n_components

        self.ncomponents = ncomponents

        variational_distribution = gpytorch.variational.\
                NaturalVariationalDistribution(inducing_points.size(-2), \
                batch_shape=torch.Size([ncomponents]))
        variational_strategy = IndependentMultitaskVariationalStrategy(\
                VariationalStrategy(self,inducing_points, \
                variational_distribution, learn_inducing_locations=True), \
                num_tasks=ncomponents)

        super(GPModel, self).__init__(variational_strategy)

        self.mean_module = gpytorch.means.ConstantMean(batch_shape=\
                torch.Size([ncomponents]))

        # Set up the GPs.

        self.covar_module = gpytorch.kernels.ScaleKernel(\
                gpytorch.kernels.MaternKernel(nu=1.5, \
                ard_num_dims=len(keys), \
                active_dims=list(range(len(keys))), \
                batch_shape=torch.Size([ncomponents])), \
                batch_shape=torch.Size([ncomponents]))

        # Set initial values.

        if load:
            self.load_state_dict(torch.load('gp.pth'))
            self.eval()
        else:
            self.covar_module.base_kernel.lengthscale = torch.reshape(\
                    torch.tensor([[[(config.parameters[k]["limits"][1] - \
                    config.parameters[k]["limits"][0])/5 \
                    for k in keys]] for i in range(ncomponents)]).float(),\
                    [ncomponents, 1, len(keys)])

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

################################################################################
#
# Create a function to do the Gaussian Process fit.
#
################################################################################

def run_gp_fit():
    # Load in the data.

    x, y, ncomponents = load_data(grid="train", pca=True)

    # Set up the data as tensors.

    train_x = torch.from_numpy(x).float()
    train_y = torch.from_numpy(y[:,:ncomponents]).float()
    print(train_y.size(-1), ncomponents)

    # Create the Gaussian process.

    induce = numpy.random.choice(numpy.arange(train_x.size(0)), \
            size=min(500,train_x.size(0)), replace=False)

    inducing_points = train_x[induce, :].unsqueeze(0).repeat(ncomponents, 1, 1)

    model = GPModel(inducing_points=inducing_points, ncomponents=ncomponents)
    likelihood = gpytorch.likelihoods.MultitaskGaussianLikelihood(\
            num_tasks=ncomponents, rank=0, has_global_noise=False, \
            noise_constraint=gpytorch.constraints.GreaterThan(1e-3**2))

    likelihood.task_noises = 0.1**2

    # If a GPU is available, put the data there.

    if torch.cuda.is_available():
        inducing_points = inducing_points.cuda()
        train_x = train_x.cuda()
        train_y = train_y.cuda()
        model = model.cuda()
        likelihood = likelihood.cuda()

    train_dataset = TensorDataset(train_x, train_y)
    train_loader = DataLoader(train_dataset, batch_size=1000, shuffle=True)

    # Find optimal hyperparameters

    model.train()
    likelihood.train()

    # Use the adam optimizer

    hyperparameter_optimizer = torch.optim.Adam([\
            {'params': model.hyperparameters()},\
            {'params': likelihood.parameters()}], lr=1.0e-2)

    variational_ngd_optimizer = gpytorch.optim.NGD(\
            model.variational_parameters(), num_data=train_y.size(0), lr=0.1)

    # "Loss" for GPs - the marginal log likelihood

    mll = gpytorch.mlls.VariationalELBO(likelihood, model, \
            num_data=train_y.size(0))

    num_epochs = 3000

    parameters = []
    for i in range(num_epochs):
        for x_batch, y_batch in train_loader:
            variational_ngd_optimizer.zero_grad()
            hyperparameter_optimizer.zero_grad()

            output = model(x_batch)

            loss = -mll(output, y_batch)
            loss.backward()

            variational_ngd_optimizer.step()
            hyperparameter_optimizer.step()

        print('Iter %d/%d - Loss: %.3f' % (i + 1, num_epochs, loss.item()))

        """
        # Store the sequence of steps for each parameter.

        labels = ["Loss"] + ["lengthscale_{0:s}".format(key) for key in \
                keys] + ["outputscale","constant","noise"]

        for icomp in range(ncomponents):
            parameters.append([loss.item(), *tuple(model.covar_module.\
                    base_kernel.lengthscale[icomp].detach().cpu().numpy().\
                    flatten()), \
                    model.covar_module.outputscale[icomp].cpu().item()**0.5, \
                    model.mean_module.constant[icomp].cpu().item(), \
                    likelihood.task_noises[icomp].cpu().item()**0.5])

        # Every 500 steps make a plot of how the parameters have been 
        # changing.
        if (i + 1) % 3000 == 0:
            for icomp in range(ncomponents):
                fig, axes = plt.subplots(nrows=4, ncols=5, figsize=(12.8,9.6))

                for j, ax in enumerate(axes.flatten()):
                    if j >= numpy.array(parameters).shape[1]:
                        ax.set_axis_off()
                        continue

                    ax.plot(numpy.arange(i+1), numpy.array(parameters)\
                            [icomp::ncomponents,j])

                    ax.set_xlabel("Steps")
                    ax.set_ylabel(labels[j].replace("_",""))

                fig.tight_layout()

                fig.savefig("stepsplot{0:d}.png".\
                        format(icomp))
                plt.close(fig)
                plt.clf()
        """

    # Bring the data and models back from the GPU, if they were there.

    if torch.cuda.is_available():
        inducing_points = inducing_points.cpu()
        train_x = train_x.cpu()
        train_y = train_y.cpu()
        model = model.cpu()
        likelihood = likelihood.cpu()

    # Finally, save the model.

    torch.save(model.state_dict(), "gp.pth")
    torch.save(likelihood.state_dict(), "likelihood.pth")
    torch.save(inducing_points, 'inducing_points.pt')

################################################################################
#
# Test the GP fit against 1D slices.
#
################################################################################

def test_in_1d():
    # Directory to read results from.

    fitdir = "gpfit_plots"

    if not os.path.exists(fitdir):
        os.mkdir(fitdir)

    # Load in the data all at once and store it.

    x = {}
    y = {}

    for i, key in enumerate(keys):
        # Load in the data.

        x[key], y[key], ncomponents = load_data(grid="1d", key=key, pca=True)

        #y[key] = y[key][numpy.argsort(x[key][:,i]),:]
        #x[key] = numpy.sort(x[key][:,i])
        x[key] = x[key][:,i]

    # Also the transformed data

    model = GPModel(load=True)

    likelihood = gpytorch.likelihoods.MultitaskGaussianLikelihood(\
            num_tasks=ncomponents, rank=0, has_global_noise=False)
    likelihood.load_state_dict(torch.load('likelihood.pth'))
    likelihood.eval()

    # Loop through the different parameters and plot.

    for icomp in range(ncomponents):
        fig, axes = plt.subplots(nrows=2, ncols=4, figsize=(9,4), \
                gridspec_kw=dict(left=0.06,right=0.95, top=0.95, bottom=0.1, \
                wspace=0.4, hspace=0.25))

        for ikey, (key, ax) in enumerate(zip(keys, axes.flatten())):
            # Generate a predicted curve and uncertainties.

            model.eval()
            likelihood.eval()

            with torch.no_grad(), gpytorch.settings.fast_pred_var():
                # The x-values at which to predict values from the Gaussian 
                # processes.
                x_pred = torch.from_numpy(numpy.concatenate([\
                        numpy.linspace(config.parameters[k]["limits"][0], \
                        config.parameters[k]["limits"][1], 500).\
                        reshape((-1,1)) if k == key else \
                        numpy.repeat(config.parameters[k]["value"], 500).\
                        reshape((-1,1)) for k in keys], axis=1)).float()

                observed_pred = likelihood(model(x_pred))

                lower, upper = observed_pred.confidence_region()

            # Now make a plot of the best fit Gaussian process

            ax.errorbar(x[key], y[key][:,icomp], fmt=".k", capsize=0)

            ax.plot(x_pred[:,ikey].numpy(), observed_pred.mean[:,icomp].\
                    numpy(), 'k', alpha=0.5)
            ax.fill_between(x_pred[:,ikey].numpy(), lower[:,icomp].numpy(), \
                    upper[:,icomp].numpy(), color="k", alpha=0.25)

            ax.set_xlim(config.parameters[key]["limits"][0], \
                    config.parameters[key]["limits"][1])

            ax.set_xlabel("$"+key+"$")
            ax.set_ylabel("$w_{0:d}$".format(icomp+1))

        fig.savefig(fitdir+"/component{0:d}_1Dproj.png".format(icomp))

################################################################################
#
# Test the GP fit against the full parameter space test data.
#
################################################################################

def test_gp_full(plot=False):
    # Also the transformed data

    x_grid, y_grid, ncomponents = load_data(grid="test", pca=True)
    residuals = y_grid.copy()

    # Also load in the Gaussian process fits.

    model = GPModel(load=True)

    # Subtract off the mean from the original data.

    test_x = torch.from_numpy(x_grid).float()
    test_y = torch.from_numpy(y_grid).float()

    test_dataset = TensorDataset(test_x, test_y)
    test_loader = DataLoader(test_dataset, batch_size=1000, shuffle=False)

    with torch.no_grad():
        index = 0
        for x_batch, y_batch in test_loader:
            preds = model(x_batch)
            means = preds.mean

            residuals[index:index+1000,:model.ncomponents] -= means.numpy()

            index += 1000

    icomp = 0

    print(residuals[:,icomp].min(), residuals[:,icomp].max(), \
            residuals[:,icomp].std(), astropy.stats.mad_std(residuals[:,icomp]))
    print(y_grid[:,icomp].max() - y_grid[:,icomp].min(), numpy.abs(\
            residuals[:,icomp]).max() / (y_grid[:,icomp].max() - \
            y_grid[:,icomp].min()))

    print(numpy.argmax(residuals[:,icomp]), \
            x_grid[numpy.argmax(residuals[:,icomp]),:])
    print(numpy.argmin(residuals[:,icomp]), \
            x_grid[numpy.argmin(residuals[:,icomp]),:])

    if plot:
        # Create a figure to put the results in.

        fig = plt.figure()
        ax = plt.axes(projection='3d')

        # Now plot the residuals.

        ax.scatter3D(x_grid[:,0], x_grid[:,1], x_grid[:,2], \
                c=residuals[:,icomp])

        plt.show()

################################################################################
#
# Learn the next point at which to fill parameter space with.
#
################################################################################

def learn_next_point():
    # Also load in the Gaussian process fits.

    model = GPModel(load=True)

    # Subtract off the mean from the original data.

    parameters, x_grid = generate_samples(10000, mode="random")
    y_grid = numpy.random.uniform(size=(10000,model.ncomponents))

    test_x = torch.from_numpy(x_grid).float()
    test_y = torch.from_numpy(y_grid).float()
    variance = torch.from_numpy(y_grid).float()

    test_dataset = TensorDataset(test_x, test_y)
    test_loader = DataLoader(test_dataset, batch_size=1000, shuffle=False)

    with torch.no_grad():
        index = 0
        for x_batch, y_batch in test_loader:
            preds = model(x_batch)

            variance[index:index+1000,:model.ncomponents] = preds.variance

            index += 1000

    # Now plot the residuals.

    max_variance_index = torch.argmax(variance.sum(axis=1))

    return parameters[max_variance_index:max_variance_index+1]

################################################################################
#
# Here's where the code can take actions.
#
################################################################################

if (args.action == 'make'):
    grid = get_directory(grid=args.grid)

    if not os.path.exists(grid):
        os.mkdir(grid)

    os.chdir(grid)

    if grid == "1DGrids/":
        for key in keys:
            if not os.path.exists(key):
                os.mkdir(key)

            os.chdir(key)

            parameters, samples = generate_samples(config.nsamples_1d, \
                    mode="LHS", grid_keys=[key])
            numpy.save("parameters.npy", parameters)

            os.chdir("..")
    elif grid == "TestGrid/":
        parameters, samples = generate_samples(config.nsamples_test, \
                mode="random")
        numpy.save("parameters.npy", parameters)
    else:
        parameters, samples = generate_samples(config.nsamples_train, \
                mode="LHS")
        numpy.save("parameters.npy", parameters)

    os.chdir("..")

elif (args.action == 'run'):
    if "1" in args.grid:
        for key in keys:
            run_models(grid=args.grid, key=key)
    else:
        run_models(grid=args.grid)

elif (args.action == 'pca'):
    run_pca_decomposition(plot=True)

elif (args.action == 'gpfit'):
    run_gp_fit()

elif (args.action == "test"):
    if "1" in args.grid:
        test_in_1d()
    elif args.grid == "test":
        test_gp_full(plot=True)

elif (args.action == 'learn'):
    directory = get_directory(grid="train")

    run_pca_decomposition(plot=False)
    run_gp_fit()

    for i in range(args.nlearn):
        # Do stuff

        new_parameters = learn_next_point()
        print(new_parameters)

        # Update the parameters.

        old_parameters = numpy.load(directory+"parameters.npy", \
                allow_pickle=True)

        parameters = numpy.concatenate((old_parameters, new_parameters))

        numpy.save(directory+"parameters.npy", parameters)

        # Run the new model to add it to the grid.

        run_models(grid="train")

        # Re-run the PCA and GP fits to update the surrogate model.

        run_pca_decomposition(plot=False)
        run_gp_fit()



from .libinterferometry import Visibilities
import numpy

def concatenate(visibilities):

    for i, vis in enumerate(visibilities):
        if i == 0:
            u = vis.u.copy()
            v = vis.v.copy()
            real = vis.real.copy()
            imag = vis.imag.copy()
            amp = vis.amp.copy()
            weights = vis.weights.copy()
            freq = vis.freq.copy()
            if type(vis.baseline) != type(None):
                baseline = vis.baseline.copy()
                incl_baselines = True
            else:
                incl_baselines = False
        else:
            u = numpy.concatenate((u, vis.u.copy()))
            v = numpy.concatenate((v, vis.v.copy()))
            real = numpy.concatenate((real, vis.real.copy()))
            imag = numpy.concatenate((imag, vis.imag.copy()))
            amp = numpy.concatenate((amp, vis.amp.copy()))
            weights = numpy.concatenate((weights, vis.weights.copy()))
            if incl_baselines:
                if type(vis.baseline) != type(None):
                    baseline = numpy.concatenate((baseline,vis.baseline.copy()))
                else:
                    incl_baselines = False

    if incl_baselines:
        return Visibilities(u, v, freq, real, imag, weights, baseline=baseline)
    else:
        return Visibilities(u, v, freq, real, imag, weights)



#!/usr/bin/env python3

from ..constants.physics import c, m_p, G
from ..constants.physics import k as k_b
from ..constants.astronomy import M_sun, AU, arcsec
from .YSOModel import YSOModel
from .. import interferometry as uv
from .. import spectroscopy as sp
from .. import misc
from .. import dust
from .. import gas
import scipy.signal
import tempfile
import numpy
import time
import os

################################################################################
#
# Create a function which returns a model of the data.
#
################################################################################

def run_flared_model(visibilities, params, parameters, plot=False, ncpus=1, \
        source="flared", plot_vis=False, nice=None, ftcode="galario"):

    # Set the values of all of the parameters.

    p = {}
    for key in parameters:
        if parameters[key]["fixed"]:
            if parameters[key]["value"] in parameters.keys():
                if parameters[parameters[key]["value"]]["fixed"]:
                    value = parameters[parameters[key]["value"]]["value"]
                else:
                    value = params[parameters[key]["value"]]
            else:
                value = parameters[key]["value"]
        else:
            value = params[key]

        if key[0:3] == "log":
            p[key[3:]] = 10.**value
        else:
            p[key] = value

    # Make sure alpha and beta are defined.

    if p["disk_type"] in ["exptaper","dartois-exptaper"]:
        t_rdisk = p["T0"] * (p["R_disk"] / 1.)**-p["q"]
        p["h_0"] = ((k_b*(p["R_disk"]*AU)**3*t_rdisk) / (G*p["M_star"]*M_sun * \
                p["mu"]*m_p))**0.5 / AU
    else:
        p["h_0"] = ((k_b * AU**3 * p["T0"]) / (G*p["M_star"]*M_sun * \
                p["mu"]*m_p))**0.5 / AU
    p["beta"] = 0.5 * (3 - p["q"])
    p["alpha"] = p["gamma"] + p["beta"]

    # Set up the dust.

    dustopac = p["dust_file"]

    dust_gen = dust.DustGenerator(dust.__path__[0]+"/data/"+dustopac)

    ddust = dust_gen(p["a_max"] / 1e4, p["p"])
    edust = dust_gen(1.0e-4, 3.5)

    # Set up the gas.

    gases = []
    abundance = []
    freezeout = []

    index = 1
    while index > 0:
        if "gas_file"+str(index) in p:
            g = gas.Gas()
            g.set_properties_from_lambda(gas.__path__[0]+"/data/"+\
                    p["gas_file"+str(index)])

            gases.append(g)
            abundance.append(p["abundance"+str(index)])
            freezeout.append(p["freezeout"+str(index)])

            index += 1
        else:
            index = -1

    # Make sure we are in a temp directory to not overwrite anything.

    original_dir = os.environ["PWD"]
    temp_dir = tempfile.TemporaryDirectory()
    os.chdir(temp_dir.name)

    # Write the parameters to a text file so it is easy to keep track of them.

    f = open("params.txt","w")
    for key in p:
        f.write("{0:s} = {1}\n".format(key, p[key]))
    f.close()

    # Set up the model. 

    m = YSOModel()
    m.add_star(mass=p["M_star"], luminosity=p["L_star"],temperature=p["T_star"])

    if "dartois" in p["disk_type"]:
        ntheta = 101
    else:
        ntheta = 51

    if "ulrich" in p["envelope_type"]:
        p["R_grid"] = 20*p["R_env"]
    else:
        p["R_grid"] = max(5*p["R_disk"],300)

    if ftcode == "trift":
        nr = int(numpy.ceil((numpy.log10(p["R_grid"]) - -1.) / 0.04))
        p["R_in_grid"] = 0.1
    else:
        nr = 100
        p["R_in_grid"] = p["R_in"]
    m.set_spherical_grid(p["R_in_grid"], p["R_grid"], nr, ntheta, 2, \
            code="radmc3d")

    if p["disk_type"] == "exptaper":
        m.add_pringle_disk(mass=p["M_disk"], rmin=p["R_in"], rmax=p["R_disk"], \
                plrho=p["alpha"], h0=p["h_0"], plh=p["beta"], dust=ddust, \
                t0=p["T0"], plt=p["q"], gas=gases, abundance=abundance,\
                aturb=p["a_turb"], gamma_taper=p["gamma_taper"])
    elif p["disk_type"] == "dartois-exptaper":
        m.add_dartois_pringle_disk(mass=p["M_disk"], rmin=p["R_in"], \
                rmax=p["R_disk"], plrho=p["alpha"], h0=p["h_0"], plh=p["beta"],\
                dust=ddust, tmid0=p["Tmid0"], tatm0=p["Tatm0"], zq0=p["zq0"], \
                pltgas=p["pltgas"], delta=p["delta"], gas=gases, \
                abundance=abundance, freezeout=freezeout, aturb=p["a_turb"], \
                gamma_taper=p["gamma_taper"])
    elif p["disk_type"] == "dartois-truncated":
        m.add_dartois_disk(mass=p["M_disk"], rmin=p["R_in"], rmax=p["R_disk"], \
                plrho=p["alpha"], h0=p["h_0"], plh=p["beta"], dust=ddust, \
                tmid0=p["Tmid0"], tatm0=p["Tatm0"], zq0=p["zq0"], \
                pltgas=p["pltgas"], delta=p["delta"], gas=gases, \
                abundance=abundance, freezeout=freezeout, aturb=p["a_turb"])
    else:
        m.add_disk(mass=p["M_disk"], rmin=p["R_in"], rmax=p["R_disk"], \
                plrho=p["alpha"], h0=p["h_0"], plh=p["beta"], dust=ddust, \
                t0=p["T0"], plt=p["q"], gas=gases, abundance=abundance,\
                aturb=p["a_turb"])

    if p["envelope_type"] == "ulrich":
        m.add_ulrich_envelope(mass=p["M_env"], rmin=p["R_in"], rmax=p["R_env"],\
                cavpl=p["ksi"], cavrfact=p["f_cav"], dust=edust, \
                t0=p["T0_env"], tpl=p["q_env"], gas=gases, abundance=abundance,\
                aturb=p["a_turb_env"], rcent=p["R_c"])
    elif p["envelope_type"] == "ulrich-extended":
        m.add_ulrichextended_envelope(mass=p["M_env"], rmin=p["R_in"], \
                rmax=p["R_env"], rcent=p["R_c"], cavpl=p["ksi"], \
                cavrfact=p["f_cav"], theta_open=p["theta_open"], \
                zoffset=p["zoffset"], t0=p["T0_env"], tpl=p["q_env"], \
                aturb=p["a_turb_env"], dust=edust, gas=gases, \
                abundance=abundance)
    elif p["envelope_type"] == "ulrich-tapered":
        m.add_tapered_ulrich_envelope(mass=p["M_env"], rmin=p["R_in"], \
                rmax=p["R_env"], gamma=p["gamma_env"], cavpl=p["ksi"], \
                cavrfact=p["f_cav"], dust=edust, t0=p["T0_env"], \
                tpl=p["q_env"], gas=gases, abundance=abundance,\
                aturb=p["a_turb_env"], rcent=p["R_c"])
    elif p["envelope_type"] == "ulrich-tapered-extended":
        m.add_tapered_ulrichextended_envelope(mass=p["M_env"], rmin=p["R_in"], \
                rmax=p["R_env"], gamma=p["gamma_env"], rcent=p["R_c"], \
                cavpl=p["ksi"], cavrfact=p["f_cav"], \
                theta_open=p["theta_open"], zoffset=p["zoffset"], \
                t0=p["T0_env"], tpl=p["q_env"], aturb=p["a_turb_env"], \
                dust=edust, gas=gases, abundance=abundance)
    else:
        pass

    m.grid.set_wavelength_grid(0.1,1.0e5,500,log=True)

    # Run the images/visibilities/SEDs.

    for j in range(len(visibilities["file"])):
        # If sub-velocity resolution is requested, adjust the frequencies.

        if visibilities["subsample"][j] * visibilities["averaging"][j] > 1:
            dfreq = visibilities["data"][j].freq[1:] - \
                    visibilities["data"][j].freq[0:-1]
            freq = []
            for i in range(0, visibilities["data"][j].freq.size):
                freq.append(visibilities["data"][j].freq[i] + dfreq[i%\
                        (visibilities["data"][j].freq.size-1)] * \
                        (numpy.linspace(1./(visibilities["subsample"][j]*\
                        visibilities["averaging"][j]*2), 1.-1/(\
                        visibilities["subsample"][j]*\
                        visibilities["averaging"][j]*2), \
                        visibilities["subsample"][j]*\
                        visibilities["averaging"][j]) - 0.5))

            freq = numpy.concatenate(freq)
        else:
            freq = visibilities["data"][j].freq

        # Shift the wavelengths by the velocities.

        b = p["v_sys"]*1.0e5 / c
        lam = c / freq / 1.0e-4
        wave = lam * numpy.sqrt((1. - b) / (1. + b))

        # Set the wavelengths for RADMC3D to use.

        m.set_camera_wavelength(wave)

        if ftcode == "galario":
            if p["docontsub"]:
                m.run_image(name=visibilities["lam"][j], nphot=1e5, \
                        npix=visibilities["npix"][j], lam=None, \
                        pixelsize=visibilities["pixelsize"][j], \
                        tgas_eq_tdust=True, scattering_mode_max=0, \
                        incl_dust=True, incl_lines=True, loadlambda=True, \
                        incl=p["i"], pa=p["pa"], dpc=p["dpc"], \
                        code="radmc3d", verbose=False, \
                        writeimage_unformatted=True, setthreads=ncpus, \
                        nice=nice)

                m.run_image(name="cont", nphot=1e5, \
                        npix=visibilities["npix"][j], lam=None, \
                        pixelsize=visibilities["pixelsize"][j], \
                        tgas_eq_tdust=True, scattering_mode_max=0, \
                        incl_dust=True, incl_lines=False, loadlambda=True, \
                        incl=p["i"], pa=p["pa"], dpc=p["dpc"], \
                        code="radmc3d", verbose=False, \
                        writeimage_unformatted=True, setthreads=ncpus, \
                        nice=nice)

                m.images[visibilities["lam"][j]].image -= m.images["cont"].image
            else:
                m.run_image(name=visibilities["lam"][j], nphot=1e5, \
                        npix=visibilities["npix"][j], lam=None, \
                        pixelsize=visibilities["pixelsize"][j], \
                        tgas_eq_tdust=True, scattering_mode_max=0, \
                        incl_dust=False, incl_lines=True, loadlambda=True, \
                        incl=p["i"], pa=p["pa"], dpc=p["dpc"], \
                        code="radmc3d", verbose=False, \
                        writeimage_unformatted=True, setthreads=ncpus, \
                        nice=nice)
        else:
            if p["docontsub"]:
                m.run_image(name=visibilities["lam"][j], nphot=1e5, lam=None, \
                        tgas_eq_tdust=True, scattering_mode_max=0, \
                        incl_dust=True, incl_lines=True, loadlambda=True, \
                        incl=p["i"], pa=p["pa"], dpc=p["dpc"], \
                        code="radmc3d", verbose=False, \
                        writeimage_unformatted=True, setthreads=ncpus, \
                        nice=nice, unstructured=True, nostar=True)

                m.run_image(name="cont", nphot=1e5, lam=None, \
                        tgas_eq_tdust=True, scattering_mode_max=0, \
                        incl_dust=True, incl_lines=False, loadlambda=True, \
                        incl=p["i"], pa=p["pa"], dpc=p["dpc"], \
                        code="radmc3d", verbose=False, \
                        writeimage_unformatted=True, setthreads=ncpus, \
                        nice=nice, unstructured=True, nostar=True)

                m.images[visibilities["lam"][j]].image -= m.images["cont"].image
            else:
                m.run_image(name=visibilities["lam"][j], nphot=1e5, lam=None, \
                        tgas_eq_tdust=True, scattering_mode_max=0, \
                        incl_dust=False, incl_lines=True, loadlambda=True, \
                        incl=p["i"], pa=p["pa"], dpc=p["dpc"], \
                        code="radmc3d", verbose=False, \
                        writeimage_unformatted=True, setthreads=ncpus, \
                        nice=nice, unstructured=True, nostar=True)

        # Extinct the data, if included.

        velocity = c * (float(visibilities["freq"][j])*1.0e9 - freq)/ \
                (float(visibilities["freq"][j])*1.0e9) / 1.0e5

        tau = p["tau0"] * numpy.exp(-(velocity - p["v_ext"])**2 / \
                (2*p["sigma_vext"]**2))

        extinction = numpy.exp(-tau)

        if ftcode == "galario":
            for i in range(len(m.images[visibilities["lam"][j]].freq)):
                m.images[visibilities["lam"][j]].image[:,:,i,:] *= extinction[i]
        else:
            for i in range(len(m.images[visibilities["lam"][j]].freq)):
                m.images[visibilities["lam"][j]].image[:,i] *= extinction[i]

        # And then if sub-sampling, sum the image along the frequency axis to
        # get back to the right size, and also Hanning smooth.

        if visibilities["subsample"][j] * visibilities["averaging"][j] > 1 or \
                visibilities["hanning"][j]:
            # Using regular images.
            if ftcode == "galario":
                recombined = numpy.empty((visibilities["npix"][j], \
                        visibilities["npix"][j], visibilities["data"][j].freq.\
                        size*visibilities["averaging"][j],1))
                for i in range(freq.size // visibilities["subsample"][j]): 
                    recombined[:,:,i,0] = m.images[visibilities["lam"][j]].\
                            image[:,:,i*visibilities["subsample"][j]:(i+1)*\
                            visibilities["subsample"][j],0].mean(axis=2)

                # Now do the Hanning smoothing.

                if visibilities["hanning"][j]:
                    hanning_window = numpy.hanning(5) / numpy.hanning(5).sum()

                    recombined = scipy.signal.fftconvolve(recombined, \
                            hanning_window.reshape((1,1,hanning_window.size,\
                            1)), axes=2, mode="same")

                # Finally, average by the binning.

                binned = numpy.empty((visibilities["npix"][j], \
                        visibilities["npix"][j], visibilities["data"][j].freq.\
                        size, 1))
                for i in range(visibilities["data"][j].freq.size): 
                    binned[:,:,i,0] = recombined[:,:,i*\
                            visibilities["averaging"][j]:(i+1)*\
                            visibilities["averaging"][j],0].mean(axis=2)
            # Using unstructured images.
            else:
                recombined = numpy.empty((m.images[visibilities["lam"][j]].x.\
                        size, visibilities["data"][j].freq.size*\
                        visibilities["averaging"][j]))
                for i in range(freq.size // visibilities["subsample"][j]): 
                    recombined[:,i] = m.images[visibilities["lam"][j]].\
                            image[:,i*visibilities["subsample"][j]:(i+1)*\
                            visibilities["subsample"][j]].mean(axis=1)

                # Now do the Hanning smoothing.

                if visibilities["hanning"][j]:
                    hanning_window = numpy.hanning(5) / numpy.hanning(5).sum()

                    recombined = scipy.signal.fftconvolve(recombined, \
                            hanning_window.reshape((1,hanning_window.size)), \
                            axes=1, mode="same")

                # Finally, average by the binning.

                binned = numpy.empty((m.images[visibilities["lam"][j]].x.size, \
                        visibilities["data"][j].freq.size))
                for i in range(visibilities["data"][j].freq.size): 
                    binned[:,i] = recombined[:,i*visibilities["averaging"]\
                            [j]:(i+1)*visibilities["averaging"][j]].mean(axis=1)

            m.images[visibilities["lam"][j]].image = binned
            m.images[visibilities["lam"][j]].freq = visibilities["data"][j].freq

        # Invert to get the visibilities.

        m.visibilities[visibilities["lam"][j]] = uv.interpolate_model(\
                visibilities["data"][j].u, visibilities["data"][j].v, \
                visibilities["data"][j].freq, \
                m.images[visibilities["lam"][j]], dRA=p["x0"], dDec=p["y0"], \
                nthreads=ncpus, code=ftcode, nxy=visibilities["npix"][j], \
                dxy=visibilities["pixelsize"][j])

        if plot:
            # If sub-velocity resolution is requested, adjust the frequencies.

            if visibilities["subsample"][j] * visibilities["averaging"][j] > 1:
                dfreq = visibilities["image"][j].freq[1:] - \
                        visibilities["image"][j].freq[0:-1]
                freq = []
                for i in range(0, visibilities["image"][j].freq.size):
                    freq.append(visibilities["image"][j].freq[i] + dfreq[i%\
                            (visibilities["image"][j].freq.size-1)] * \
                            (numpy.linspace(1./(visibilities["subsample"][j]*\
                            visibilities["averaging"][j]*2), 1.-1/(\
                            visibilities["subsample"][j]*\
                            visibilities["averaging"][j]*2), \
                            visibilities["subsample"][j]*\
                            visibilities["averaging"][j]) - 0.5))

                freq = numpy.concatenate(freq)
            else:
                freq = visibilities["image"][j].freq

            # Get the wavelengths.

            lam = c / freq / 1.0e-4
            wave = lam * numpy.sqrt((1. - b) / (1. + b))

            m.set_camera_wavelength(wave)

            if p["docontsub"]:
                m.run_image(name=visibilities["lam"][j], nphot=1e5, \
                        npix=visibilities["image_npix"][j], lam=None, \
                        pixelsize=visibilities["image_pixelsize"][j], \
                        tgas_eq_tdust=True, scattering_mode_max=0, \
                        incl_dust=True, incl_lines=True, loadlambda=True, \
                        incl=p["i"], pa=p["pa"], dpc=p["dpc"], \
                        code="radmc3d", verbose=False, setthreads=ncpus, \
                        nice=nice)

                m.run_image(name="cont", nphot=1e5, \
                        npix=visibilities["image_npix"][j], lam=None, \
                        pixelsize=visibilities["image_pixelsize"][j], \
                        tgas_eq_tdust=True, scattering_mode_max=0, \
                        incl_dust=True, incl_lines=False, loadlambda=True, \
                        incl=p["i"], pa=p["pa"], dpc=p["dpc"], \
                        code="radmc3d", verbose=False, setthreads=ncpus, \
                        nice=nice)

                m.images[visibilities["lam"][j]].image -= m.images["cont"].image
            else:
                m.run_image(name=visibilities["lam"][j], nphot=1e5, \
                        npix=visibilities["image_npix"][j], lam=None, \
                        pixelsize=visibilities["image_pixelsize"][j], \
                        tgas_eq_tdust=True, scattering_mode_max=0, \
                        incl_dust=False, incl_lines=True, loadlambda=True, \
                        incl=p["i"], pa=p["pa"], dpc=p["dpc"], \
                        code="radmc3d", verbose=False, setthreads=ncpus, \
                        nice=nice)

            # Extinct the data, if included.

            velocity = c * (float(visibilities["freq"][j])*1.0e9 - freq) / \
                    (float(visibilities["freq"][j])*1.0e9) / 1.0e5

            tau = p["tau0"] * numpy.exp(-(velocity - p["v_ext"])**2 / \
                    (2*p["sigma_vext"]**2))

            extinction = numpy.exp(-tau)

            for i in range(len(m.images[visibilities["lam"][j]].freq)):
                m.images[visibilities["lam"][j]].image[:,:,i,:] *= extinction[i]

            # And then if sub-sampling, sum the image along the frequency axis 
            # to get back to the right size, and also Hanning smooth.

            if visibilities["subsample"][j] * visibilities["averaging"][j] > 1 \
                    or visibilities["hanning"][j]:
                recombined = numpy.empty((visibilities["image_npix"][j], \
                        visibilities["image_npix"][j], \
                        visibilities["image"][j].freq.size*\
                        visibilities["averaging"][j],1))
                for i in range(freq.size // visibilities["subsample"][j]): 
                    recombined[:,:,i,0] = m.images[visibilities["lam"][j]].\
                            image[:,:,i*visibilities["subsample"][j]:(i+1)*\
                            visibilities["subsample"][j],0].mean(axis=2)

                # Now do the Hanning smoothing.

                if visibilities["hanning"][j]:
                    hanning_window = numpy.hanning(5) / numpy.hanning(5).sum()

                    recombined = scipy.signal.fftconvolve(recombined, \
                            hanning_window.reshape((1,1,hanning_window.size,\
                            1)), axes=2, mode="same")

                # Finally, average by the binning.

                binned = numpy.empty((visibilities["image_npix"][j], \
                        visibilities["image_npix"][j], \
                        visibilities["image"][j].freq.size, 1))
                for i in range(visibilities["image"][j].freq.size): 
                    binned[:,:,i,0] = recombined[:,:,i * \
                            visibilities["averaging"][j]:(i+1)*\
                            visibilities["averaging"][j],0].mean(axis=2)

                m.images[visibilities["lam"][j]].image = binned
                m.images[visibilities["lam"][j]].freq = \
                        visibilities["image"][j].freq

            # Now convolve with the beam.

            x, y = numpy.meshgrid(numpy.linspace(-256,255,512), \
                    numpy.linspace(-256,255,512))

            beam = misc.gaussian2d(x, y, 0., 0., \
                    visibilities["image"][j].header["BMAJ"]/2.355/\
                    visibilities["image"][j].header["CDELT2"], \
                    visibilities["image"][j].header["BMIN"]/2.355/\
                    visibilities["image"][j].header["CDELT2"], \
                    (90-visibilities["image"][j].header["BPA"])*\
                    numpy.pi/180., 1.0)

            for ind in range(len(m.images[visibilities["lam"][j]].freq)):
                m.images[visibilities["lam"][j]].image[:,:,ind,0] = \
                        scipy.signal.fftconvolve(\
                        m.images[visibilities["lam"][j]].image[:,:,ind,0], \
                        beam, mode="same")

            if plot_vis:
                m.visibilities[visibilities["lam"][j]+"1D"] = \
                        uv.average(m.visibilities[visibilities["lam"][j]], \
                        gridsize=20, radial=True, log=True, \
                        logmin=m.visibilities[visibilities["lam"][j]].uvdist[\
                        numpy.nonzero(m.visibilities[visibilities["lam"][j]].\
                        uvdist)].min()*0.95, logmax=m.visibilities[\
                        visibilities["lam"][j]].uvdist.max()*1.05, \
                        mode="spectralline")

    os.system("rm params.txt")
    os.chdir(original_dir)

    return m



from numpy import exp,cos,sin

def gaussian2d(x,y,x0,y0,sigmax,sigmay,pa,f0):
    
    xp=(x-x0)*cos(pa)-(y-y0)*sin(pa)
    yp=(x-x0)*sin(pa)+(y-y0)*cos(pa)
    
    return f0*exp(-1*xp**2/(2*sigmax**2))*exp(-1*yp**2/(2*sigmay**2))


#!/usr/bin/env python3

from pdspy.dust import *
import numpy

dust = PAH()
dust.set_properties_from_draine("optical_constants/PAHneu_30.txt", \
        "optical_constants/PAHneu_30.txt","optical_constants/PAH_qion.txt")
dust.set_density(3.0)

amin = dust.a.min()
amax = dust.a.max()
pl = 3.5

dust.calculate_size_distribution_opacity(amin, amax, pl)

dust.write('PAH_MRN.hdf5')



import pdspy.interferometry as uv
import pdspy.imaging as im
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy

def plot_pvdiagram(visibilities, model, parameters, params, index=0, \
        plot_vis=False, image="data", contours="model", \
        model_image="beam-convolve", maxiter=100, threshold=1., uvtaper=None, \
        weighting="natural", robust=2.0, length=100, width=9, \
        image_cmap="Blues", levels=None, fontsize="medium", fig=None):
    r"""
    Plot the millimeter channel maps, along with the best fit model.

    Args:
        :attr:`visibilities` (`dict`):
            Dictionary containing the visibility data, typically as loaded by 
            :code:`utils.load_config` and :code:`utils.load_data`.
        :attr:`model` (`modeling.Model`):
            The radiative transfer model that you would like to plot the 
            visibilities of. Typically this is the output of 
            modeling.run_disk_model.
        :attr:`parameters` (`dict`):
            The parameters dictionary in the config module as loaded in by 
            :code:`utils.load_config`
        :attr:`params` (`dict`):
            The parameters of the model, typically as a dictionary mapping 
            parameter keys from the :code:`parameters` dictionary to their 
            values.
        :attr:`index` (`int`, optional):
            The visibilities dictionary typically contains a list of datasets. 
            `index` indicates which one to plot.
        :attr:`fig` (`tuple`, `(matplotlib.Figure, matplotlib.Axes)`, optional):
            If you've already created a figure and axes to put the plot in, you 
            can supply them here. Otherwise, `plot_channel_maps` will 
            generate them for you. It will use 
            :code:`visibilities["nrows"][index]` rows and 
            :code:`visibilities["ncols"][index]` columns. Default: `None`
        :attr:`image` (`str`, optional):
            Should the image show the `"data"`, `"model"`, or `"residuals"`.
            Default: `"data"`
        :attr:`contours` (`str`, optional):
            Should the image show the `"data"`, `"model"`, or `"residuals"`.
            Or could also not show contours with `"none"`. Default: `"model"`
        :attr:`model_image` (`str`, optional):
            Should the model image be made by convolving a radiatie transfer
            generated image with an estimate of the beam (`"beam-convolve"`), 
            or by generating model visibilities at the correct baselines of the
            data and then using :code:`interferometry.clean` to generate an 
            image (`"CLEAN"`). Default: `"beam-convolve"`
        :attr:`weighting` (`str`, optional):
            What weighting scheme should the model image use if 
            :code:`model_image="CLEAN"`, `"natural"`, `"robust"`, or 
            `"uniform"`. Default:`"robust"`
        :attr:`robust` (`int`, optional):
            The robust parameter when :code:`weighting="robust"`. Default: 2
        :attr:`maxiter` (`int`, optional):
            The maximum number of CLEAN iterations to perform. Default: 2
        :attr:`threshold` (`float`, optional, Jy):
            The stopping threshold for the CLEAN algorithm. Default: 0.001 Jy
        :attr:`uvtaper` (`float`, optional, lambda):
            FWHM of the Gaussian tapering to apply to the weights when running
            CLEAN. Default: None
        :attr:`length` (`int`, optional):
            Length of the box, in pixels, to use to extract the PV diagram.
            Default: `100`
        :attr:`width` (`int`, optional):
            Width of the box, in pixels, to use to extract the PV diagram. 
            Default: `100`
        :attr:`image_cmap` (`str`, optional):
            Which colormap to use for plotting the image. Default: `"Blues"`
        :attr:`levels` (`list` of `float`, optional):
            The flux levels at which to plot contours. If `None`, use 
            :code:`[0.1, 0.3, 0.5, 0.7, 0.9] x image.max()`. Default: `None`
        :attr:`fontsize` (`str` or `int`):
            What fontsize to use for labels, ticks, etc. Default: `"medium"`

    Returns:
        :attr:`fig` (`matplotlib.Figure`):
            The matplotlib figure that was used for the plot.
        :attr:`ax` (`matplotlib.Axes`):
            The matplotlib axes that were used for the plot.
        :attr:`cax` (`matplotlib.Axes`, optional):
            The matplotlib axes that were used for the colorbar, if 
            :code:`show_colorbar=True`.
    """

    # Set up the figure if none was provided.

    if fig == None:
        fig, ax = plt.subplots(nrows=visibilities["nrows"][index], \
                ncols=visibilities["ncols"][index], figsize=(10,9))
    else:
        fig, ax = fig

    # Loop through the visibilities and plot.

    for i in range(2):
        # First pass through plot the image, second pass through, plot contours.
        if i == 0:
            plot_type = image
        else:
            plot_type = contours

            # If we don't want to plot contours, just skip the second round.
            if plot_type == "none":
                continue
        
        # Get the centroid position.

        if plot_type in ["data","residuals"] or \
                model_image != "beam-convolve":
            if "x0" in params:
                x0 = visibilities["image_npix"][index]/2 - \
                        visibilities["x0"][index]/\
                        visibilities["image_pixelsize"][index]- \
                        params["x0"]/\
                        visibilities["image_pixelsize"][index]
            else:
                x0 = visibilities["image_npix"][index]/2 - \
                        visibilities["x0"][index]/\
                        visibilities["image_pixelsize"][index]- \
                        parameters["x0"]["value"]/\
                        visibilities["image_pixelsize"][index]
            if "y0" in params:
                y0 = visibilities["image_npix"][index]/2 + \
                        visibilities["y0"][index]/\
                        visibilities["image_pixelsize"][index] + \
                        params["y0"]/\
                        visibilities["image_pixelsize"][index]
            else:
                y0 = visibilities["image_npix"][index]/2 + \
                        visibilities["y0"][index]/\
                        visibilities["image_pixelsize"][index] + \
                        parameters["y0"]["value"]/\
                        visibilities["image_pixelsize"][index]
        else:
            if model_image == "beam-convolve":
                x0 = int(round(visibilities["image_npix"][index]/2))
                y0 = int(round(visibilities["image_npix"][index]/2))

        # Also grab the position angle.

        pa = (180. - (params["pa"]-90)) * numpy.pi / 180.

        # Get the correct image/contours for plotting.

        if plot_type == "data":
            plot_image = im.extract_pv_diagram(visibilities["image"][index], \
                    xy=(x0,y0), pa=pa, length=length, width=width)
        elif plot_type == "model":
            if model_image == "beam-convolve":
                model_image = model.images[visibilities["lam"][index]]
                model_image.header = visibilities["image"][index].header
            else:
                model.visibilities[visibilities["lam"][index]].weights = \
                    visibilities["data"][index].weights

                model_image = uv.clean(\
                        model.visibilities[visibilities["lam"][index]], \
                        imsize=visibilities["image_npix"][index], \
                        pixel_size=visibilities["image_pixelsize"][index], \
                        weighting=weighting, robust=robust, \
                        convolution="expsinc", mfs=False, \
                        mode="spectralline", maxiter=maxiter, \
                        threshold=threshold, uvtaper=uvtaper)[0]
                model_image.header = visibilities["image"][index].header

            # Extract the PV data.

            plot_image = im.extract_pv_diagram(model_image, xy=(x0,y0), \
                    pa=pa, length=length, width=width)
        elif plot_type == "residuals":
            residuals = Visibilities(visibilities["data"][index].u, \
                    visibilities["data"][index].v, \
                    visibilities["data"][index].freq, \
                    visibilities["data"][index].real.copy(), \
                    visibilities["data"][index].imag.copy(),\
                    visibilities["data"][index].weights)

            residuals.real -= model.visibilities[visibilities["lam"]\
                    [index]].real
            residuals.imag -= model.visibilities[visibilities["lam"]\
                    [index]].imag

            residuals_image = uv.clean(residuals, \
                    imsize=visibilities["image_npix"][index], \
                    pixel_size=visibilities["image_pixelsize"][index], \
                    weighting=weighting, robust=robust, \
                    convolution="expsinc", mfs=False, mode="spectralline", \
                    maxiter=0, uvtaper=uvtaper)[0]

            residuals_image.header = visibilities["image"][index].header

            # Extract the PV data.

            plot_image = im.extract_pv_diagram(residuals_image, xy=(x0,y0), \
                    pa=pa, length=length, width=width)

        if i == 0:
            # Plot the data.

            ax.imshow(plot_image.image[0,:,:,0].T, origin="lower", \
                    interpolation="nearest", cmap=image_cmap)

            # Transform the axes appropriately.

            dx = plot_image.x[1] - plot_image.x[0]
            dv = (plot_image.velocity[1] - plot_image.velocity[0])/1e5

            nx = plot_image.image.shape[1]
            nv = plot_image.image.shape[2]

            transform_x = ticker.FuncFormatter(Transform(0, nx, dx, '%.0f"'))
            transform_y = ticker.FuncFormatter(Transform(0, nv, dv, '%.0f'))

            ticks_x = numpy.array([-6,-3,0,3,6])
            ticks_y = numpy.array([-3,-2,-1,0,1,2,3])

            ax.set_xticks(nx/2+ticks_x/dx+0.5)
            ax.set_yticks(nv/2+ticks_y/dv-0.5)
            ax.get_xaxis().set_major_formatter(transform_x)
            ax.get_yaxis().set_major_formatter(transform_y)

            # Add axes labels.

            ax.set_xlabel('Offset', fontsize=fontsize, labelpad=10)
            ax.set_ylabel('Velocity [km s$^{-1}$]', fontsize=fontsize, \
                    labelpad=10)

            ax.tick_params(axis="both", which="both", labelsize=fontsize)

            # Set the aspect ratio correctly.

            ax.set_aspect(nx/nv)

            # And the axes range.

            ax.set_xlim(0, nx-1)
            ax.set_ylim(0, nv-1)

        else:
            # Contour the model data.

            if levels is None:
                levels = (numpy.arange(5)+0.5)/5. * plot_image.image.max()

            ax.contour(plot_image.image[0,:,:,0].T, colors="black", \
                    levels=levels)

    return fig, ax

# Define a useful class for plotting.

class Transform:
    def __init__(self, xmin, xmax, dx, fmt):
        self.xmin = xmin
        self.xmax = xmax
        self.dx = dx
        self.fmt = fmt

    def __call__(self, x, p):
        return self.fmt% ((x-(self.xmax-self.xmin+1)/2)*self.dx)




from .plot_SED import plot_SED
from .plot_continuum_image import plot_continuum_image
from .plot_1D_visibilities import plot_1D_visibilities
from .plot_2D_visibilities import plot_2D_visibilities
from .plot_scattered_light import plot_scattered_light
from .plot_channel_maps import plot_channel_maps
from .plot_pvdiagram import plot_pvdiagram
from . import colormaps
from .Transform import Transform
from .cubeshow import cubeshow



from ..constants.astronomy import arcsec
from ..table import Table
import numpy
import scipy.ndimage.filters
import scipy.ndimage.morphology
import scipy.optimize
import scipy.signal
import matplotlib.pyplot as plt
import os
import astropy
import astropy.coordinates

def find(image, threshold=5, include_radius=20, window_size=40, \
        source_list=None, list_search_radius=1.0, list_threshold=5, \
        beam=[1.0,1.0,0.0], user_aperture=False, aperture=15, \
        fit_aperture=15, include_flux_unc=False, flux_unc=0.1, \
        bootstrap_unc=True, output_plots=None, just_find=False):

    # If plots of the fits have been requested, make the directory if it 
    # doesn't already exist.

    if output_plots != None:
        os.system("rm -r "+output_plots)
        os.mkdir(output_plots)

    # Find potential sources in the image.

    base_image = image.image[:,:,0,0]

    neighborhood = scipy.ndimage.morphology.generate_binary_structure(2,2)

    local_max = scipy.ndimage.filters.maximum_filter(base_image, \
            footprint=neighborhood) == base_image

    background = (base_image == 0)

    eroded_background = scipy.ndimage.morphology.binary_erosion(background, \
            structure=neighborhood, border_value=1)

    detected_peaks = local_max - eroded_background

    potential_sources = numpy.column_stack(numpy.nonzero(detected_peaks))
    
    # If we are providing a source list, the threshold should be lower than
    # for a blind search. Throw away for good any sources that don't meet
    # this threshold.

    if type(source_list) != type(None):
        for coords in potential_sources:
            if image.image[coords[0], coords[1], 0, 0] < list_threshold * \
                    image.unc[coords[0], coords[1], 0, 0]:
                detected_peaks[coords[0], coords[1]] = 0.

        potential_sources = numpy.column_stack(numpy.nonzero(detected_peaks))

    # First, throw away any potential source that does not meet the
    # threshold cut requirement.

    for coords in potential_sources:
        if image.image[coords[0], coords[1], 0, 0] < threshold * \
                image.unc[coords[0], coords[1], 0, 0]:
            detected_peaks[coords[0], coords[1]] = 0.

    # Next, if a source list was provided for searching purposes, throw away
    # any objects not within the requisite radius from a listed source.

    if type(source_list) != type(None):
        coords = astropy.coordinates.SkyCoord(source_list["ra"].tolist(), \
                source_list["dec"].tolist(), frame='icrs')

        pixcoords = image.wcs.wcs_world2pix(coords.ra.degree, \
                coords.dec.degree, 1, ra_dec_order=True)

        arcsec_in_pixels = arcsec / (abs(image.wcs.wcs.cdelt[0]) * numpy.pi/180)

        for peak_coords in potential_sources:
            for count, pixcoord in enumerate(zip(pixcoords[0], pixcoords[1])):
                if (numpy.sqrt((pixcoord[0] - peak_coords[1])**2 + \
                               (pixcoord[1] - peak_coords[0])**2) <\
                               0.5 * list_search_radius * arcsec_in_pixels):
                    #detected_peaks[peak_coords[0], peak_coords[1]] = 1.
                    if image.image[peak_coords[0], peak_coords[1], 0, 0] > \
                            list_threshold * image.unc[peak_coords[0], \
                            peak_coords[1], 0, 0]:
                        detected_peaks[peak_coords[0], peak_coords[1]] = 1.
                        break
                #elif (count == len(pixcoords[0])-1):
                #    detected_peaks[peak_coords[0], peak_coords[1]] = 0.

    potential_sources = numpy.column_stack(numpy.nonzero(detected_peaks))

    # Search for potential sources that are probably the same source and
    # make sure we're only finding it once.

    good = numpy.repeat(True, len(potential_sources))
    for i in range(len(potential_sources)):
        for j in range(len(potential_sources)):
            if (j != i) and good[i] and good[j]:
                coords = potential_sources[i]
                coords2 = potential_sources[j]

                d = numpy.sqrt( (coords[0] - coords2[0])**2 + \
                        (coords[1] - coords2[1])**2 )

                if (d < include_radius) and (d > 0):
                    inbetween = bresenham_line(coords[0],coords[1],coords2[0], \
                            coords2[1])

                    for k, coords3 in enumerate(inbetween):
                        if image.image[coords3[0],coords3[1], 0, 0] < 2.0 * \
                                image.unc[coords3[0], coords3[1], 0 ,0]:
                            break

                        if k == len(inbetween)-1:
                            if image.image[coords[0], coords[1]] > \
                                    image.image[coords2[0], coords2[1]]:
                                detected_peaks[coords2[0], coords2[1]] = 0.
                                good[j] = False
                            else:
                                detected_peaks[coords[0], coords[1]] = 0.
                                good[i] = False

    potential_sources = numpy.column_stack(numpy.nonzero(detected_peaks))

    if just_find:
        return potential_sources

    # Now we have a good list of detected sources. Fit all of them with a
    # Gaussian to measure positions and fluxes.

    sources = []

    for coords in potential_sources:
        # Set up the parameter guesses for fitting.

        half_window = window_size / 2

        xmin = max(0,coords[1]-half_window)
        xmax = min(coords[1]+half_window,image.image.shape[1])
        ymin = max(0,coords[0]-half_window)
        ymax = min(coords[0]+half_window,image.image.shape[1])

        x, y = numpy.meshgrid(numpy.linspace(xmin, xmax-1, xmax - xmin), \
                numpy.linspace(ymin, ymax-1, ymax - ymin))

        z = image.image[int(ymin):int(ymax),int(xmin):int(xmax),0,0]
        sigma_z = image.unc[int(ymin):int(ymax),int(xmin):int(xmax),0,0]

        beam_to_sigma = arcsec / (abs(image.wcs.wcs.cdelt[0]) * numpy.pi/180) /\
                2.355

        xc, yc = coords[1], coords[0]
        params = numpy.array([xc, yc, beam[0]*beam_to_sigma, \
                beam[1]*beam_to_sigma, beam[2], image.image[yc,xc,0,0]])

        bm = [image.header["BMAJ"]/abs(image.wcs.wcs.cdelt[0])/2.355,\
                image.header["BMIN"]/abs(image.wcs.wcs.cdelt[0])/2.355,\
                image.header["BPA"]*numpy.pi/180.]

        # Fit the source with a Gaussian.

        try:
            p, sigma_p = fit_source(coords, x, y, z, sigma_z, params, \
                    bootstrap_unc=bootstrap_unc, beam=bm)
        except ValueError:
            continue

        # Create a new source to add.

        new_source = numpy.empty((16,), dtype=p.dtype)
        new_source[0:12][0::2] = p[0:6]
        new_source[0:12][1::2] = sigma_p[0:6]

        # Before doing aperture photometry, find any sources within a provided 
        # radius and fit them so they can be subtracted out of the sky
        # subtraction window.

        nsources = 1

        for coords2 in potential_sources:
            d = numpy.sqrt( (coords[0] - coords2[0])**2 + \
                    (coords[1] - coords2[1])**2 )

            if (d < include_radius) and (d > 0):
                xc, yc = coords2[1], coords2[0]
                params = numpy.array([xc, yc, beam[0]*beam_to_sigma, \
                        beam[1]*beam_to_sigma, beam[2], image.image[yc,xc,0,0]])

                try:
                    new_p, sigma_new_p = fit_source(coords2, x, y, z, sigma_z, \
                            params, bootstrap_unc=False)
                except ValueError:
                    continue

                p = numpy.hstack([p, new_p])
                nsources += 1

        # Do some aperture photometry for the source.

        if nsources > 1:
            new_z = z.copy() - gaussian2d(x, y, p[6:], nsources-1)
        else:
            new_z = z.copy()

        if not user_aperture:
            aperture = 3 * numpy.sqrt(new_source[4]*new_source[6])

        try:
            sky = numpy.median(new_z[numpy.logical_and(\
                    numpy.sqrt((coords[1]-x)**2 + (coords[0]-y)**2) > aperture,\
                    numpy.sqrt((coords[1]-x)**2 + (coords[0]-y)**2) <= \
                    4*aperture)])
        except IndexError:
            sky = -1.0e-5
            print("Error in source:", len(sources))

        new_source[12] = image.image[coords[0], coords[1], 0, 0] - sky
        new_source[13] = image.unc[coords[0], coords[1], 0, 0]
        new_source[14] = (new_z[numpy.sqrt((new_source[0]-x)**2 + \
                (new_source[2]-y)**2) < aperture] - sky).sum()
        new_source[15] = numpy.sqrt((sigma_z[numpy.sqrt((new_source[0]-x)**2+ \
                (new_source[2]-y)**2) < aperture]**2).sum())

        # Add the newly found source to the list of sources.

        sources.append(new_source)

        # Plot the image slice.

        if output_plots != None:
            fig, ax = plt.subplots(nrows=2, ncols=2)

            ax[0,0].set_title("Data")
            ax[0,0].imshow(z, origin="lower", interpolation="nearest", \
                    vmin=z.min(), vmax=z.max())
            ax[0,1].set_title("Uncertainty")
            ax[0,1].imshow(sigma_z, origin="lower", interpolation="nearest", \
                    vmin=z.min(), vmax=z.max())
            ax[1,0].set_title("Model")

            ax[1,0].imshow(gaussian2d(x, y, p, nsources), origin="lower", \
                    interpolation="nearest", vmin=z.min(), vmax=z.max())
            ax[1,1].set_title("Data-Model")
            ax[1,1].imshow(z - gaussian2d(x, y, p, nsources), \
                    origin="lower", interpolation="nearest", vmin=z.min(), \
                    vmax=z.max())

            circle1 = plt.Circle((new_source[0] - coords[1] + half_window, \
                    new_source[2] - coords[0] + half_window), aperture, \
                    edgecolor='r', facecolor="none")
            ax[0,0].add_artist(circle1)
            circle2 = plt.Circle((new_source[0] - coords[1] + half_window, \
                    new_source[2] - coords[0] + half_window), \
                    4*aperture, edgecolor='r', facecolor="none")
            ax[0,0].add_artist(circle2)

            fig.savefig(output_plots+"/source_{0:d}.pdf".format(len(sources)-1))

            plt.close(fig)

    if len(sources) > 0:
        sources = Table(numpy.array(sources), names=("x", \
                "x_unc","y","y_unc","sigma_x","sigma_x_unc","sigma_y", \
                "sigma_y_unc","pa", "pa_unc","f",'f_unc',"Peak_Flux", \
                "Peak_Flux_unc","Flux","Flux_unc"))

    if hasattr(image, "wcs"):
        try:
            ra, dec = image.wcs.wcs_pix2world(sources["x"], sources["y"], 1)

            temp = astropy.coordinates.SkyCoord(ra, dec, unit='deg')

            sources['ra'] = temp.ra.to_string(unit=astropy.units.hour)
            sources['dec'] = temp.dec.to_string()

            sources['ra_unc'] = abs(image.wcs.wcs.cdelt[0]) * sources['x_unc'] / \
                    (180. / numpy.pi) / arcsec
            sources['dec_unc'] = abs(image.wcs.wcs.cdelt[1]) * sources['y_unc'] / \
                    (180. / numpy.pi) / arcsec

            sources['FWHM_x'] = 2.35482 * abs(image.wcs.wcs.cdelt[0]) * \
                    sources['sigma_x'] / (180. / numpy.pi) / arcsec
            sources['FWHM_y'] = 2.35482 * abs(image.wcs.wcs.cdelt[1]) * \
                    sources['sigma_y'] / (180. / numpy.pi) / arcsec
            sources['FWHM_x_unc'] = 2.35482 * abs(image.wcs.wcs.cdelt[0]) * \
                    sources['sigma_x_unc'] / (180. / numpy.pi) / arcsec
            sources['FWHM_y_unc'] = 2.35482 * abs(image.wcs.wcs.cdelt[1]) * \
                    sources['sigma_y_unc'] / (180. / numpy.pi) / arcsec

            sources['flux'] = sources['f'] * sources['sigma_x'] * \
                    sources['sigma_y'] * 2 * numpy.pi
            sources['flux_unc'] = 2*numpy.pi * numpy.sqrt( \
                    (sources['f_unc']*sources['sigma_x']*sources['sigma_y'])**2+\
                    (sources['f']*sources['sigma_x_unc']*sources['sigma_y'])**2+\
                    (sources['f']*sources['sigma_x'] * sources['sigma_y_unc'])**2)

            if "BMAJ" in image.header:
                beam_per_pixel = abs(image.wcs.wcs.cdelt.prod()) / \
                        (numpy.pi*image.header['BMAJ']*image.header['BMIN']/ \
                        (4*numpy.log(2)))

                sources['flux'] *= beam_per_pixel
                sources['flux_unc'] *= beam_per_pixel
                sources['Flux'] *= beam_per_pixel
                sources['Flux_unc'] *= beam_per_pixel

            if include_flux_unc:
                sources['flux_unc'] = numpy.sqrt(sources['flux_unc']**2 + \
                        (flux_unc * sources['flux'])**2)
                sources['Flux_unc'] = numpy.sqrt(sources['Flux_unc']**2 + \
                        (flux_unc * sources['Flux'])**2)
        except:
            print(sources["x"], sources["y"], temp)

    return sources

def fit_source(coords, x, y, z, sigma_z, params, bootstrap_unc=True, \
        beam=None):
    # Try a least squares fit.

    func = lambda p, n, x, y, z, sigma: \
            ((z - gaussian2d(x, y, p, n)) / sigma).reshape((z.size,))

    # Try to automatically determine the aperture.

    z3 = z.copy()
    z3[z/sigma_z < 2.] = 0.0

    xm, ym = x[0, int(x.shape[1]/2)], y[int(y.shape[0]/2), 0]

    for i in range(z.shape[0]):
        for j in range(z.shape[1]):
            if z3[i,j] != 0.:
                inbetween = bresenham_line(int(coords[0]-ym+z.shape[0]/2), \
                        int(coords[1] - xm + z.shape[1]/2),i,j)

                for k, coords3 in enumerate(inbetween):
                    if z3[coords3[0],coords3[1]] < 2.0 * \
                            sigma_z[coords3[0], coords3[1]]:
                        z3[i,j] = 0.0

    params, cov, infodict, mesg, ier = scipy.optimize.leastsq(func, \
            params, args=(1, x, y, z3, sigma_z), full_output=True)

    fit_aperture = 3 * numpy.sqrt(abs(params[2]*params[3]))

    # Only fit to pixels in a certain aperture around the source. This is 
    # because otherwise the fitting routine has a habit of finding other 
    # peaks or negative dips when fitting faint sources.

    for i in range(1):
        z2 = numpy.zeros(z.shape)
        sigma_z2 = numpy.zeros(z.shape) + 1.0e20

        if i > 0:
            params[0:6] = p[0:6]
            ap = 3 * numpy.sqrt(abs(params[2]*params[3]))
        else:
            ap = fit_aperture

        z2[numpy.sqrt((x - params[0])**2 + (y - params[1])**2) < ap] = \
                z[numpy.sqrt((x - params[0])**2 + (y - params[1])**2) < ap]
        sigma_z2[numpy.sqrt((x - params[0])**2 + (y - params[1])**2) < ap]=\
                sigma_z[numpy.sqrt((x - params[0])**2 + (y - params[1])**2)\
                < ap]

        p, cov, infodict, mesg, ier = scipy.optimize.leastsq(func, \
                params, args=(1, x, y, z2, sigma_z2), full_output=True)

    if (type(cov) == type(None)):
        raise ValueError('The fit failed')

    # Fix the least squares result for phi because it seems to like to go
    # crazy.

    p[4] = numpy.fmod(numpy.fmod(p[4], numpy.pi)+numpy.pi, numpy.pi)

    # Take the absolute values of sigma_x and sigma_y in case they 
    # managed to go negative.

    p[2:4] = abs(p[2:4])

    # Now calculate the uncertainty on the parameters using bootstrapping.

    if bootstrap_unc:
        s_res = sigma_z.mean()
        ps = []

        for i in range(100):
            randomDelta = interferometer_noise(z.shape, s_res, beam)
            randomdataZ = z + randomDelta
            randomdataSigma_Z = sigma_z.copy()

            randomdataZ[numpy.sqrt((x-params[0])**2 + (y-params[1])**2) >= \
                    ap] = 0.
            randomdataSigma_Z[numpy.sqrt((x - params[0])**2 + \
                    (y - params[1])**2) >= ap] = 1.0e20

            randomfit, randomcov = scipy.optimize.leastsq(func, params, \
                    args=(1, x, y, randomdataZ, randomdataSigma_Z), \
                    full_output=False)

            ps.append(randomfit)

        ps = numpy.array(ps)

        pfit_bootstrap = numpy.mean(ps,0)
        perr_bootstrap = numpy.std(ps,0)

        pfit_bootstrap[4] = numpy.fmod(numpy.fmod(pfit_bootstrap[4], \
                numpy.pi)+numpy.pi, numpy.pi)

        return p, perr_bootstrap

    # Calculate the uncertainties on the parameters.

    else:
        sigma_p = numpy.sqrt(numpy.diag((func(p, 1, x, y, z2, \
                sigma_z2)**2).sum()/(y.size - p.size) * cov))

        return p, sigma_p

def interferometer_noise(shape, sigma, beam):

    test = numpy.random.normal(0.0, sigma, shape[0]*shape[1]).reshape(shape)

    x, y = numpy.meshgrid(numpy.arange(shape[1]),numpy.arange(shape[0]))

    gaus = gaussian2d(x, y, [shape[1]/2, shape[0]/2, beam[0], beam[1], \
            beam[2], 1])

    new_test = scipy.signal.fftconvolve(test, gaus, mode='same')
    new_test *= sigma / new_test.std()

    return new_test

def gaussian2d(x, y, params, n=1):

    model = numpy.zeros(x.shape)

    for i in range(n):
        x0, y0, sigmax, sigmay, pa, f0 = tuple(params[i*6:i*6+6])

        xp=(x-x0)*numpy.sin(pa)-(y-y0)*numpy.cos(pa)
        yp=(x-x0)*numpy.cos(pa)+(y-y0)*numpy.sin(pa)

        model += f0*numpy.exp(-1*xp**2/(2*sigmax**2))* \
                numpy.exp(-1*yp**2/(2*sigmay**2))

    return model

def bresenham_line(x,y,x2,y2):
    """Brensenham line algorithm"""
    steep = 0
    coords = []
    dx = abs(x2 - x)
    if (x2 - x) > 0: sx = 1
    else: sx = -1
    dy = abs(y2 - y)
    if (y2 - y) > 0: sy = 1
    else: sy = -1
    if dy > dx:
        steep = 1
        x,y = y,x
        dx,dy = dy,dx
        sx,sy = sy,sx
    d = (2 * dy) - dx
    for i in range(0,dx):
        if steep: coords.append((y,x))
        else: coords.append((x,y))
        while d >= 0:
            y = y + sy
            d = d - (2 * dx)
        x = x + sx
        d = d + (2 * dy)
    coords.append((x2,y2))
    return coords
